# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015-2017, Ignacio Fdez. Galván                        *
#***********************************************************************

from __future__ import (unicode_literals, division, absolute_import, print_function)

from pyparsing import ParseException, __version__ as pyparsing_version
from emil_grammar import EMIL_Grammar
from abstract_flow import *
from re import match, search, IGNORECASE, compile as re_compile

# set of precompiled regular expressions
re_export = re_compile(r'export\s+(\S*)\s*=\s*(.*?)\s*$', flags=IGNORECASE)
re_eval = re_compile(r'eval\s+(\S*)\s*=\s*(.*?)\s*$', flags=IGNORECASE)
re_exit = re_compile(r'exit(\s+(\d+|\$\w*)?)?\s*$', flags=IGNORECASE)
re_shell = re_compile(r'(?:shell|unix)\s+(\S.*?)\s*$', flags=IGNORECASE)
re_exec = re_compile(r'exec\s+(\S.*?)\s*$', flags=IGNORECASE)
re_include = re_compile(r'include\s+(\S.*?)\s*$', flags=IGNORECASE)
re_dowhile = re_compile(r'do\s*while\s*$', flags=IGNORECASE)
re_enddo = re_compile(r'end\s*do\s*$', flags=IGNORECASE)
re_foreach = re_compile(r'foreach\s+(\w+)\s+in\s+\(\s*(.*?)\s*\)\s*$', flags=IGNORECASE)
re_endforeach = re_compile(r'end\s*foreach\s*$', flags=IGNORECASE)
re_dogeo = re_compile(r'do\s*geo\s*$', flags=IGNORECASE)
re_if = re_compile(r'if\s*\(\s*(.*?)\s*(=+|!=|ne|-file)\s*(\S.*?)\s*\)\s*$', flags=IGNORECASE)
re_ifgoto = re_compile(r'if\s*\(\s*(.*?)\s*(=+|!=|ne|-file)\s*(\S.*?)\s*\)\s*go\s*to\s+(\S+)\s*$', flags=IGNORECASE)
re_endif = re_compile(r'end\s*if\s*$', flags=IGNORECASE)
re_echo = re_compile(r'echo\s+(\S.*?)\s*$', flags=IGNORECASE)
re_file = re_compile(r'file\s+(\S.*?)\s*$', flags=IGNORECASE)
re_rm = re_compile(r'rm\s+(-?force\s+)?(\S.*?)\s*$', flags=IGNORECASE)
re_copy = re_compile(r'(copy|save|clone|collect)\s+(-?force\s+)?(\S*)\s+(\S*)\s*$', flags=IGNORECASE)
re_label = re_compile(r'label\s+(\S+)\s*$', flags=IGNORECASE)
re_goto = re_compile(r'go\s*to\s+(\S+)\s*$', flags=IGNORECASE)

class EMILException(Exception):
  pass

# TODO: check for "ln" once (if) LINK is supported
def check_commandline(item, commandline, forbidden=None):
  name = match(r'\S*', item).group(0).upper()
  if (forbidden is not None):
    if (search(forbidden, commandline)):
      raise EMILException('Forbidden characters in "{0}" command'.format(name))
  if (match(r'cp\s', commandline)):
    raise EMILException('Use COPY instead of "{0} cp"'.format(name))
  if (match(r'rm\s', commandline)):
    raise EMILException('Use RM instead of "{0} rm"'.format(name))
  if (match(r'export\s', commandline)):
    raise EMILException('Use EXPORT instead of "{0} export"'.format(name))

################################################################################
# Parse the result of an initial EMIL_Grammar run into abstract flow tokens
################################################################################

def EMIL_Parse(input_file):
  '''Parse an input file in EMIL language, returns a tuple where the first
     element is Group of abstract flow tokens, and the second element is
     a possibly empty dictionary of files embedded files (filename: content)
  '''

  # Open the file first, to ensure it is utf8-encoded
  with utf8_open(input_file, 'r') as f_in:
    try:
      items = EMIL_Grammar.parseFile(f_in, parseAll=True).asList()
    except ParseException as e:
      if (pyparsing_version >= '2.0.2'):
        raise EMILException('Parsing error: {0}\n{1}'.format(e, e.markInputline()))
      else:
        raise EMILException('Parsing error: {0}'.format(e))

  blocks = [[]]
  files = {}
  level = 0
  leveltype = [['main']]

  # Process each parsed item, adding the corresponding "Statement" subclass
  # to the token list
  for item in items:

    if (item[0] == '&'):
      # Remove final empty lines (empty lines in the middle could be significant)
      while (item[-1] == ''):
        del item[-1]
      blocks[level].append(Program(item))

    elif (item[0] == '>'):

      # TODO: variant/option for binary files?
      # >>> FILE
      # this is not added to the token list, but returned as a separate dict
      # so the wrapper can create the files before running the tokens
      if (re_file.match(item[1])):
        re_match = re_file.match(item[1])
        filename = re_match.group(1)
        files[filename]=item[2]

      # >>> EXIT
      elif (re_exit.match(item[1])):
        re_match = re_exit.match(item[1])
        val = re_match.group(2)
        blocks[level].append(Break(val))

      # >>> EXPORT
      elif (re_export.match(item[1])):
        re_match = re_export.match(item[1])
        blocks[level].append(Assignment(re_match.group(1), re_match.group(2)))

      # >>> EVAL
      elif (re_eval.match(item[1])):
        re_match = re_eval.match(item[1])
        blocks[level].append(Assignment(re_match.group(1), re_match.group(2), literal=False))

      # >>> UNIX
      # >>> SHELL
      elif (re_shell.match(item[1])):
        re_match = re_shell.match(item[1])
        check_commandline(item[1], re_match.group(1))
        blocks[level].append(System(re_match.group(1)))

      # >>> EXEC
      elif (re_exec.match(item[1])):
        re_match = re_exec.match(item[1])
        check_commandline(item[1], re_match.group(1), r'[&`;>\\]')
        blocks[level].append(System(re_match.group(1), parallel=True))

      # TODO: variant for including code inside a module
      # >>> INCLUDE
      elif (re_include.match(item[1])):
        re_match = re_include.match(item[1])
        blocks[level].append(Include(re_match.group(1)))

      # >>> DO WHILE
      elif (re_dowhile.match(item[1])):
        blocks.append([])
        level += 1
        leveltype.append(['dowhile'])
      # >>> END DO
      elif (re_enddo.match(item[1])):
        if ((leveltype[level][0] != 'dowhile') and (leveltype[level][0] != 'foreach')):
          raise EMILException('Grouping error')
        # Do not allow empty DO loops
        if (len(blocks[level]) > 0):
          if (leveltype[level][0] == 'foreach'):
            var = leveltype[level][1]
            val = leveltype[level][2]
          else:
            var = None
            val = None
          for item in blocks[level]:
            item.level = level
          blocks[level-1].append(Group(blocks[level], rerun=True, var=var, val=val))
        level -= 1
        blocks.pop()
        leveltype.pop()

      # >>> IF
      # >>> IF () GOTO
      elif (re_if.match(item[1]) or re_ifgoto.match(item[1])):
        re_match = re_if.match(item[1]) or re_ifgoto.match(item[1])
        var = re_match.group(1)
        op = re_match.group(2).lower()
        val = re_match.group(3)
        name = None
        if (re_match.lastindex > 3):
          name = re_match.group(4)
        if (var.lower() == 'iter'):
          var = '$MOLCAS_ITER'
        if (op == '-file'):
          if (var != ''):
            raise EMILException('Error parsing IF command')
          test = Expression(var, 'file', val)
        elif ((op == 'ne') or (op == '!=')):
          test = Expression(var, 'ne', val)
        else:
          test = Expression(var, 'eq', val)
        if (name):
          item = Jump(name)
          item.level = level+1
          blocks[level].append(Group([item], condition=test))
        else:
          blocks.append([])
          level += 1
          leveltype.append(['if', test])
      # >>> END IF
      elif (re_endif.match(item[1])):
        if (leveltype[level][0] != 'if'):
          raise EMILException('Grouping error')
        if (len(blocks[level]) > 0):
          test = leveltype[level][1]
          for item in blocks[level]:
            item.level = level
          blocks[level-1].append(Group(blocks[level], condition=test))
        level -= 1
        blocks.pop()
        leveltype.pop()

      # >>> ECHO (undocumented)
      elif (re_echo.match(item[1])):
        re_match = re_echo.match(item[1])
        blocks[level].append(Setting('echo', re_match.group(1).lower()))

      # >>> RM
      elif (re_rm.match(item[1])):
        re_match = re_rm.match(item[1])
        blocks[level].append(ParTask('rm', (re_match.group(1) is not None), [re_match.group(2)]))
        
      # >>> COPY, SAVE, CLONE, COLLECT
      elif (re_copy.match(item[1])):
        re_match = re_copy.match(item[1])
        blocks[level].append(ParTask(re_match.group(1).lower(), (re_match.group(2) is not None), [re_match.group(3), re_match.group(4)]))
        
      # >>> FOREACH
      elif (re_foreach.match(item[1])):
        re_match = re_foreach.match(item[1])
        var, val = re_match.groups()
        blocks.append([])
        level += 1
        leveltype.append(['foreach', var, val])
      # >>> END FOREACH (for consistency)
      elif (re_endforeach.match(item[1])):
        if (leveltype[level][0] != 'foreach'):
          raise EMILException('Grouping error')
        # Allow empty foreach loops
        #if (len(blocks[level]) > 0):
        if (True):
          var = leveltype[level][1]
          val = leveltype[level][2]
          for item in blocks[level]:
            item.level = level
          blocks[level-1].append(Group(blocks[level], rerun=True, var=var, val=val))
        level -= 1
        blocks.pop()
        leveltype.pop()

      # >>> LABEL
      elif (re_label.match(item[1])):
        re_match = re_label.match(item[1])
        blocks[level].append(Label(re_match.group(1)))

      # >>> GOTO (extension)
      elif (re_goto.match(item[1])):
        re_match = re_goto.match(item[1])
        blocks[level].append(Jump(re_match.group(1)))

      # TODO:
      # >>> LINK
      # >>> DO GEO
      # >>> VERBATIM
      # >>> ELSE (extension)
      # >>> IF ... > / IF ... < (extension)

      # >>> DO GEO unsupported for now
      elif (re_dogeo.match(item[1])):
        raise EMILException('"DO GEO" is unsupported')

      else:
        raise EMILException('Unknown or unsupported statement:\n  > {0}'.format(item[1]))

  if (level != 0):
    raise EMILException('Grouping error')

  return Group(blocks[0]), files



# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015-2020, Ignacio Fdez. Galván                        *
#***********************************************************************

from __future__ import (unicode_literals, division, absolute_import, print_function)
try:
  from builtins import bytes
except ImportError:
  from future.builtins import bytes
try:
  from six import text_type
except ImportError:
  text_type = str

from os import environ, access, W_OK, X_OK, listdir, remove, getpid, getcwd, makedirs, symlink, devnull
from os.path import isfile, isdir, isabs, join, basename, splitext, getmtime, abspath, exists, relpath, realpath
from datetime import datetime
from shutil import copy2, move, rmtree, Error
from subprocess import check_output, STDOUT, CalledProcessError
from re import compile as re_compile, search, sub, MULTILINE, IGNORECASE
from io import BytesIO
from resource import getrusage, RUSAGE_CHILDREN
from glob import glob
from errno import EEXIST, ENOENT
from shlex import split as shsplit
from contextlib import contextmanager
from textwrap import fill

from tee import teed_call
from emil_parse import EMIL_Parse, EMILException
from python_parse import Python_Parse
from molcas_aux import *
from check_test import *
from validate import *

# python2 has no FileNotFoundError, so we will have to check
# the errno attribute of the raised exception
try:
  #python3
  FileNotFoundError
except:
  #python2 (IOError when opening files, OSError when removing files)
  FileNotFoundError = (IOError, OSError)

hcbanner = '''#
      &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
      &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
      &&&                                                      &&&
      &&&                                                      &&&
      &&&    OpenMolcas source code or broken installation     &&&
      &&&                                                      &&&
      &&&                                                      &&&
      &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
      &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
 \n
            version $V$
        patch level $P$
                    $Px$
 \n
                 Copyright (C) The OpenMolcas Authors
           For the author list and the recommended citation,
                    consult the file CONTRIBUTORS.
'''
#===============================================================================

# TODO:
# MOLCAS_Count
# MOLCAS_EMIL_DEBUG
# MOLCAS_IN_GEO
# MOLCAS_ISDEV
# MOLCAS_LASTMOD
# MOLCAS_LINK
# MOLCAS_LOG
# MOLCAS_MOLDEN
# MOLCAS_OLDPWD
# MOLCAS_PARNELL
# MOLCAS_SERIAL
# MOLCAS_SUBMIT_DIR
# MOLCAS_SUBMIT_PWD
# MOLCAS_TASK_INPUT
# MOLCAS_TIME
# MOLCAS_UNIX_SECURE
# MOLCAS_ZOMBIE

class MolcasException(Exception):
  pass

class Molcas_wrapper(object):

  version = 'py2.13'
  rc = 0

  def __init__(self, **kwargs):
    # Get Molcas istallation
    self.molcas = abspath(get_utf8('MOLCAS', default=''))
    if (self.molcas == ''):
      raise MolcasException('"MOLCAS" is not defined')
    if (not isfile(join(self.molcas, '.molcashome'))):
      raise MolcasException('"{0}" is not a valid MOLCAS installation'.format(self.molcas))
    # Get location of sources (other than MOLCAS)
    self.sources = []
    self.sources.append(abspath(get_utf8('OPENMOLCAS_SOURCE', default='')))
    self.sources.append(abspath(get_utf8('MOLCAS_SOURCE', default='')))
    self.sources = [i for i in self.sources if (i != '' and i != self.molcas)]
    # Some settings
    self.allow_shell = True
    self.echo = True
    self.input_filename = ''
    self.only_validate = False
    if ('allow_shell' in kwargs):
      self.allow_shell = bool(kwargs['allow_shell'])
    if ('echo' in kwargs):
      self.echo = bool(kwargs['echo'])
    if ('input_file' in kwargs):
      self.input_filename = kwargs['input_file']
    if ('warning' in kwargs):
      self.warning = kwargs['warning']
    if ('stamp' in kwargs):
      self.stamp = kwargs['stamp']
    if ('validate' in kwargs):
      self.only_validate = kwargs['validate']
    self.licensee = None
    self.keywords = None
    self.rc = None
    self._ready = False
    self._goto = False
    self._parse_rte()
    self._parse_codes()
    self._parse_alias()
    self._parse_keywords()
    self._resources = (0,0,0)
    self._check_count = 0
    self._loop_level = 0
    self._nest_level = 0
    self._last_module = True
    self.parnell = join(self.molcas, 'bin', 'parnell.exe')
    if (self.input_filename):
      self.read_input(self.input_filename)

  @contextmanager
  def Loop(self):
    self.enter_loop()
    try:
      yield
    finally:
      self.exit_loop()

  def _parse_rte(self):
    '''Parse the molcas.rte file'''
    rte = {}
    filename = join(self.molcas, 'molcas.rte')
    try:
      r_f = utf8_open(filename, 'r')
    except IOError:
      pass
    else:
      with r_f:
        comment_match = re_compile(r'#.*$')
        rte_match = re_compile(r'(\w+)\s*=\s*([\'\"])(.*)\2')
        for line in r_f:
          line = comment_match.sub('', line)
          match = rte_match.search(line)
          if (match):
            rte[match.group(1)] = match.group(3)
    self._rte = rte

  def _parse_codes(self):
    '''Parse the list of return codes (in plain or perl syntax)'''
    code_list = {}
    filename = join(self.molcas, 'data', 'rcodes.txt')
    rc_match = re_compile(r'(\w+)\s*=\s*(\d+)')
    try:
      c_f = utf8_open(filename, 'r')
    except IOError:
      filename = join(self.molcas, 'data', 'warnings.plx')
      rc_match = re_compile(r'\$(\w+)\s*=\s*(\d+)\s*;')
      try:
        c_f = utf8_open(filename, 'r')
      except IOError:
        c_f = utf8_open(devnull, 'r')
    with c_f:
      for line in c_f:
        match = rc_match.search(line)
        if (match):
          code_list[int(match.group(2))] = match.group(1)
    self._rc_codes = code_list

  def _parse_alias(self):
    '''Parse the list of aliases (in plain or perl syntax)'''
    self.alias = {}
    filename = join(self.molcas, 'data', 'modalias.txt')
    alias_match = re_compile(r'([^#].*=.*)')
    pair_match = re_compile(r'\s*(\w*)\s*=\s*(\w*)\s*')
    try:
      a_f = utf8_open(filename, 'r')
    except IOError:
      filename = join(self.molcas, 'data', 'alias.plx')
      alias_match = re_compile(r'\s*%NickName\s*=\s*\((.*)\);')
      pair_match = re_compile(r'q\/(.*)\/=>q\/(.*)\/')
      try:
        a_f = utf8_open(filename, 'r')
      except IOError:
        return
    with a_f:
      for line in a_f:
        match = alias_match.match(line)
        if (match):
          pairs = match.group(1).split(',')
          for i in pairs:
            match = pair_match.match(i)
            if (match):
              self.alias[match.group(1)] = match.group(2)

  def _parse_keywords(self):
    try:
      self.keywords = read_db(join(self.molcas, 'data', 'keyword.xml'))
    except NameError:
      self.keywords = 'no_lxml'
    except:
      pass

  def read_environment(self):
    '''Read environment variables from rc files (in priority order)'''
    rcfiles = [dotmolcas('molcasrc'), join(self.molcas, 'molcasrc')]
    var_match = re_compile(r'\s*([^#].*?)\s*=\s*(.*)')
    for rcfile in rcfiles:
      try:
        r_f = utf8_open(rcfile, 'r')
        for line in r_f:
          match = var_match.match(line)
          if (match):
            if (match.group(1) not in environ):
              set_utf8(match.group(1), match.group(2))
        r_f.close()
      except (OSError, IOError):
        pass

  def setup(self):
    if (not self._rte):
      raise MolcasException('"{0}" is not a valid MOLCAS installation'.format(self.molcas))
    if (self._ready):
      print('*******************************')
      print('* ERROR -- setup called twice *')
      print('*******************************')
      self.rc = '_RC_GENERAL_ERROR_'
      return
    # Define some hard-coded/initial environment variables
    set_utf8('LANG', 'C')
    set_utf8('LC_ALL', '')
    set_utf8('EMIL_RC2', '0')
    set_utf8('EMIL_InLoop', '0')
    set_utf8('MOLCAS_STRUCTURE', '0')
    if (self.check_license() == 0):
      set_utf8('MOLCAS_STAMP', '5')
    else:
      set_utf8('MOLCAS_STAMP', 'UNKNOWN_VARIABLE')
    # Get current directory
    self.currdir = getcwd()
    set_utf8('CurrDir', self.currdir)
    set_utf8('MOLCAS_SUBMIT_DIR', self.currdir)
    if ((self.currdir is None) or (self.currdir == '')):
      raise MolcasException('"CurrDir" is not defined')
    # Get project name
    usepid = False
    top_project = get_utf8('MOLCAS_PROJECT', default='NAME')
    if (top_project.upper() == 'NAMEPID'):
      top_project = splitext(self.input_filename)[0]
      usepid = True
    elif (top_project.upper() == 'NAME'):
      top_project = splitext(self.input_filename)[0]
    elif (top_project.upper() == 'TIME'):
      top_project = '{:%Y_%m_%d_%H_%M}'.format(datetime.now())
    self.project = get_utf8('Project', default=top_project)
    self.project = basename(self.project)
    if ((self.project is None) or (self.project == '')):
      self.project = 'Noname'
    set_utf8('Project', self.project)
    # Get scratch directory
    self._tmpdir = get_utf8('TMPDIR', default='/tmp')
    self._pid = get_utf8('MOLCAS_PID', default=getpid())
    self.scratch = get_utf8('WorkDir', default='')
    if (self.scratch == ''):
      top_workdir = get_utf8('MOLCAS_WORKDIR', default=self._tmpdir)
      if (isabs(top_workdir)):
        self.scratch = join(top_workdir, self.project)
        if (usepid):
          self.scratch = '{0}.{1}'.format(self.scratch, self._pid)
      else:
        if (top_workdir.upper() == 'PWD'):
          self.scratch = self.currdir
        else:
          self.scratch = join(self.currdir, top_workdir)
    self.scratch = abspath(self.scratch)
    set_utf8('WorkDir',self.scratch)
    if ((self.scratch is None) or (self.scratch == '')):
      raise MolcasException('"WorkDir" is not defined')
    if (not self.only_validate):
      if (self.parallel_task(['base', self.scratch]) != 0):
        raise MolcasException('parnell failed to create a WorkDir at {0}'.format(self.scratch))
    # Get output directory
    self.output = get_utf8('MOLCAS_OUTPUT', default=get_utf8('PBS_O_WORKDIR', default=self.currdir))
    if (not isabs(self.output)):
      if (self.output.upper() == 'PWD'):
        self.output = self.currdir
      elif (self.output.upper() == 'WORKDIR'):
        self.output = self.scratch
      elif (self.output.upper() == 'NAME'):
        self.output = join(self.currdir, self.project)
      else:
        self.output = join(self.currdir, self.output)
    set_utf8('MOLCAS_OUTPUT', self.output)
    if ((self.output is None) or (self.output == '')):
      raise MolcasException('Output directory is not defined')
    if (not self.only_validate):
      try:
        makedirs(self.output)
      except OSError as e:
        if ((e.errno == EEXIST) and (isdir(self.output))):
          pass
        else:
          raise
      if (not isdir(self.output) or not access(self.output, W_OK)):
        raise MolcasException('"{0}" is not a writable directory'.format(self.output))
    # Get other variables
    if (get_utf8('MOLCAS_MEM', default='') == ''):
      set_utf8('MOLCAS_MEM', self._rte['DEFMOLCASMEM'])
    if (get_utf8('MOLCAS_DISK', default='') == ''):
      set_utf8('MOLCAS_DISK', self._rte['DEFMOLCASDISK'])
    if (get_utf8('MOLCAS_NPROCS', default='') == ''):
      set_utf8('MOLCAS_NPROCS', '1')
    if (get_utf8('SubProject', default='') == ''):
      set_utf8('SubProject', '')
    if (get_utf8('GeoDir', default='') == ''):
      set_utf8('GeoDir', join(self.scratch, self.project+'.GEO'))
    self.save_mode = get_utf8('MOLCAS_SAVE', default='repl').lower()
    if (get_utf8('MOLCAS_NEW_WORKDIR', default='NO').upper() == 'YES'):
      self.delete_scratch(True)
    self._is_empty = self._is_scratch_empty()
    self._ready = True

  def rc_to_name(self, rc):
    try:
      rc = int(rc)
      if (rc in self._rc_codes):
        return self._rc_codes[rc]
      else:
        return rc
    except:
      return rc

  @property
  def rc_num(self):
    try:
      rc = int(self.rc)
    except:
      num = [k for k,v in self._rc_codes.items() if v==self.rc]
      if (num):
        rc = int(num[0])
      else:
        rc = -1
    return rc

  def print_banner(self):
    try:
      with utf8_open(join(self.molcas, 'data', 'banner.txt'), 'r') as banner_file:
        banner = banner_file.read().rstrip('\n')
    except FileNotFoundError as e:
      if (e.errno == ENOENT):
        banner = hcbanner.rstrip('\n')
      else:
        raise
    tag = '(unknown)'
    tag_x = ''
    try:
      with utf8_open(join(self.molcas, '.molcasversion'), 'r') as version_file:
        for line in version_file:
          if (search('\.x\d', line)):
            tag_x = line.rstrip()
          else:
            tag = line.rstrip()
    except FileNotFoundError as e:
      if (e.errno == ENOENT):
        try:
          command = ["git", "describe", "--always", "--match", "v*", "--dirty"]
          line = check_output(command, stderr=STDOUT).decode('utf-8')
          if (search('\.x\d', line)):
            tag_x = line.rstrip()
          else:
            tag = line.rstrip()
        except:
          tag = '(unknown)'
      else:
        raise
    if ((tag_x != '') and (tag == '(unknown)')):
      tag = tag_x
      tag_x = ''
    v_match = re_compile('v(\d+\.\d+)[\.-](.*)')
    match = v_match.match(tag)
    if (match):
      version = match.group(1)
      patch = match.group(2)
    else:
      version = tag
      patch = ''
    match = v_match.match(tag_x)
    if (match):
      patch_x = match.group(2)
    else:
      patch_x = ''
      with utf8_open(join(self.molcas, '.molcashome')) as homefile:
        if (homefile.read().rstrip() not in ['openmolcas', '']):
          patch_x = 'x'
    date = datetime.now()
    # Remove comments and empty lines (not if there are spaces)
    banner = sub(r'#.*$', '', banner, flags=MULTILINE)
    banner = sub(r'^\n', '', banner, flags=MULTILINE)
    # Replace version and patch numbers
    patch_re = re_compile(r'^(.*)\$P\$(.*)$', flags=MULTILINE)
    patch_match = patch_re.search(banner)
    if (patch_match):
      fmt = '{{0}}\n{{1:>{0}}} {{2}}'.format(len(patch_match.group(1))-1)
      s = fmt.format(patch_match.group(0), '', '$Px$')
      if (banner.find('$Px$') < 0):
        banner = patch_re.sub(s, banner)
    banner = sub(r'\$V\$', version, banner)
    banner = sub(r'\$P\$', patch, banner)
    if (patch_x):
      banner = sub(r'\$Px\$', patch_x, banner)
    else:
      banner = sub(r'.*\$Px\$.*\n', patch_x, banner)
    # Christmas tree between Dec. 18th and Jan. 8th
    # (disabled)
    #if ((date.month == 12 and date.day > 17) or
    #    (date.month == 1 and date.day < 9)):
    #  banner = banner.replace('@', '^')
    #  banner = banner.replace('!', 'o')
    #else:
    #  banner = banner.replace('@', ' ')
    #  banner = banner.replace('!', ' ')
    # Do the main character substitution
    rep = get_utf8('MOLCAS_BANNER')
    if (patch_x == ''):
      ch0 = '&'
      ch1 = '+'
      ch2 = '*'
      if (not rep):
        rep = 'OPENMOLCAS'
    else:
      ch0 = '&'
      ch1 = '*'
      ch2 = '+'
      if (not rep):
        rep = 'MOLCAS'
    maxlen = 0
    new_banner = []
    for line in banner.split('\n'):
      maxlen = max(maxlen, len(line))
      lline = list(line.rstrip())
      j = 0
      for i in range(len(lline)):
        if ((lline[i] == ch0) or (lline[i] == ch1)):
          lline[i] = rep[j]
          j = (j+1) % len(rep)
        elif (lline[i] == ch2):
          lline[i] = ' '
      line = ''.join(lline)
      new_banner.append(line.rstrip())
    # Reverse on Apr. 1st
    if (date.month == 4 and date.day == 1):
      fmt = '{{:<{0}}}'.format(maxlen)
      new_banner = [fmt.format(line)[::-1].rstrip() for line in new_banner]
    banner = '\n'.join(new_banner)
    print(banner)
    if hasattr(self, 'stamp'):
      fmt = '{:^72}'
      print(fmt.format('').rstrip())
      print(fmt.format('*************************************************').rstrip())
      print(fmt.format('* pymolcas version {0:<28} *'.format(self.version)).rstrip())
      print(fmt.format('*   build {0}      *'.format(self.stamp)).rstrip())
      print(fmt.format('*   (after the EMIL interpreter by V. Veryazov) *').rstrip())
      print(fmt.format('*************************************************').rstrip())
    try:
      with utf8_open(join(self.molcas, 'data', 'info.txt'), 'r') as info_file:
        print(info_file.read().rstrip('\n'))
    except:
      pass

  def file_exists(self, filename):
    filetest = expandvars(filename, default='UNKNOWN_VARIABLE')
    if (not isabs(filetest)):
      filetest = join(self.scratch, filetest)
    return isfile(filetest)

  def add_files(self, files):
    for fname, content in files.items():
      filename = expandvars(fname, default='UNKNOWN_VARIABLE')
      with utf8_open(join(self.scratch, filename), 'w') as f:
        print(content, file=f)

  def read_input(self, input_file):
    self.input_filename = input_file
    self.setup()
    if (not isfile(input_file)):
      raise MolcasException('File "{0}" does not exist'.format(input_file))
    try:
      self.flow, files = Python_Parse(input_file)
    except SyntaxError:
      try:
        self.flow, files = EMIL_Parse(input_file)
      except EMILException as message:
        raise MolcasException('EMIL: {0}'.format(message))
    if (not self.only_validate):
      self.add_files(files)

  def print_input(self):
    if (get_utf8('MOLCAS_ECHO_INPUT', default='YES').upper() != 'NO'):
      foo = text_type(self.flow)
      print('++ ---------   Input file   ---------\n')
      print(foo)
      print('\n-- ----------------------------------')

  def print_environment(self):
    # These variables must always be defined
    lines = {}
    lines['Project'] = self.project
    lines['Submitted from'] = self.currdir
    lines['Scratch area'] = self.scratch
    if (self.output == self.scratch):
      lines['Save outputs to'] = 'WORKDIR'
    else:
      lines['Save outputs to'] = self.output
    lines['Molcas'] = self.molcas
    keys = ['Project',
            'Submitted from',
            'Scratch area',
            'Save outputs to',
            'Molcas']
    # Get all the other MOLCAS* environment variables
    molcas_env = {k:get_utf8(k) for k in environ.keys() if k.startswith('MOLCAS')}
    keys.extend(sorted(molcas_env.keys()))
    for key in molcas_env.keys():
      lines[key] = molcas_env[key]
    # Do not print the variables in the list
    do_not_print = ['MOLCAS_PROJECT',
                    'MOLCAS_WORKDIR',
                    'MOLCAS',
                    'MOLCAS_STAMP',
                    'MOLCAS_SUBMIT_DIR',
                    'MOLCAS_INFO',
                    'MOLCAS_TIME',
                    'MOLCAS_DISK',
                    'MOLCAS_MEM',
                    'MOLCAS_ITER',
                    'MOLCAS_OLDPWD',
                    'MOLCAS_OUTPUT',
                    'MOLCAS_TEST',
                    'MOLCAS_LOG',
                    'MOLCAS_ECHO_INPUT',
                    'MOLCAS_ISDEV',
                    'MOLCAS_SILENT_LIC']
    for key in do_not_print:
      if key in lines:
        keys.remove(key)
        del lines[key]
    # Format string and length
    maxlen = max(len(x) for x in keys)
    ini = '  |'
    fmt = ini + ' {{0:>{0}}} = {{1}}'.format(maxlen)
    fmt2 = ini + '  {{0:>{0}}}: {{1}}'.format(maxlen)
    maxlen = maxlen + max(len(x) for x in lines.values())
    # Print the environment summary
    print('   '+'-'*(maxlen+4))
    print(ini)
    for key in keys[0:5]:
      print(fmt2.format(key, lines[key]))
    print(ini)
    if (self._is_empty):
      print(ini+' Scratch area is empty')
    else:
      print(ini+' Scratch area is NOT empty')
    print(ini)
    for key in keys[5:]:
      print(fmt.format(key, lines[key]))
    print(ini)
    print('   '+'-'*(maxlen+4))

  def write_environment(self):
    molcas_env = {k:get_utf8(k) for k in environ.keys() if k.startswith('MOLCAS')}
    # Write the molcas.env file
    fmt = '{0}={1}'
    env_file = join(self.scratch, 'molcas.env')
    with utf8_open(env_file, 'w') as env:
      print('# this is a generated file! do not edit.', file=env)
      print(fmt.format('WorkDir', self.scratch), file=env)
      print(fmt.format('Project', self.project), file=env)
      print(fmt.format('SubProject', get_utf8('SubProject', '')), file=env)
      print(fmt.format('GeoDir', get_utf8('GeoDir', '')), file=env)
      for key in sorted(molcas_env.keys()):
        print(fmt.format(key, molcas_env[key]), file=env)
    self.parallel_task(['c', '1', env_file, self.scratch])

  def show_environment(self):
    molcas_env = {k:get_utf8(k) for k in environ.keys() if k.startswith('MOLCAS')}
    keys = sorted(molcas_env.keys())
    text = ''
    line = ''
    if 'MOLCAS_STAMP' in keys:
      keys.remove('MOLCAS_STAMP')
    for key in keys:
      prev_line = line
      line = '{0}={1}'.format(key, molcas_env[key])
      if (len(line) >= 60):
        if (len(prev_line) > 0):
          text += '{0}\n'.format(prev_line)
        text += '{0}\n'.format(line)
        line = ''
      elif (len(prev_line) > 0):
        text += '{0:60} {1}\n'.format(prev_line, line)
        line = ''
    text += line
    print(text)

  def _is_scratch_empty(self):
    # Check if there are any files in $WorkDir, other than those listed
    ignore = re_compile(r'(molcas|paraops|\.input$|\.inp$|\.log$|\.err$)')
    try:
      files = [f for f in listdir(self.scratch) if (not ignore.search(f))]
    except FileNotFoundError as e:
      if (e.errno == ENOENT):
        files = []
    if (len(files) == 0):
      return True
    else:
      return False

  def add_resources(self, res):
    self._resources = tuple(map(sum, zip(self._resources, res)))

  def reset_resources(self):
    self._resources = (0,0,0)

  def enter_loop(self, loop_type=None):
    self._loop_level += 1
    set_utf8('EMIL_InLoop', self._loop_level)
    if (loop_type == 'do'):
      # This simply marks if we ever enter a do loop
      set_utf8('MOLCAS_STRUCTURE', '1')

  def exit_loop(self):
    self._loop_level -= 1
    if (self._loop_level < 0):
      raise MolcasException('Bad nesting level')
    set_utf8('EMIL_InLoop', self._loop_level)

  def startup(self):
    if (self._ready):
      if (not self.only_validate):
        self.write_environment()
      print('   This run of MOLCAS is using the pymolcas driver')
      if (self.licensee):
        print('   Licensed to: {0}'.format(self.licensee))
      self.print_banner()
      print('\n')
      if (hasattr(self, 'warning') and self.warning):
        print(self.warning)
      self.print_environment()
      if (not self.only_validate):
        print('')
        self.print_input()

  def end(self):
    if (hasattr(self, 'rc')):
      # In case of error in parallel, collect the "stdout" files from the slaves,
      # since they may have important information (create an empty "stdout" file
      # to avoid further errors)
      if ((self.rc_to_name(self.rc) == '_RC_INTERNAL_ERROR_') and not self.is_serial):
        f = 'stdout'
        open(join(self.scratch, f), 'a').close()
        self.parallel_task(['c', '2', f, self.currdir], force=True)
      for l in self._final_rc(self.rc_num):
        print(l)
      if (self.rc_num == -1):
        print('\nAborting...')
    if (self._resources != (0,0,0)):
      print('    Timing: Wall={0:.2f} User={1:.2f} System={2:.2f}'.format(*self._resources))
    if (get_utf8('MOLCAS_KEEP_WORKDIR', default='YES').upper() == 'NO'):
      self.delete_scratch()

  def _final_rc(self, rc):
    rc_form = re_compile('rc={0}\s(.*)'.format(rc))
    text = []
    try:
      with utf8_open(join(self.molcas, 'data', 'landing.txt'), 'r') as l:
        for line in l:
          match = rc_form.match(line)
          if match:
            text.append(match.group(1))
    except:
      pass
    if not text:
      text.append('Non-zero return code')
    if text:
      maxlen = max(len(l) for l in text)
      line = '.{}.'.format('#'*(maxlen+4))
      text = ['', line] + ['.# {} #.'.format(l) for l in text] + [line, '']
    return text

  def validate(self):
    import abstract_flow
    if (self.only_validate):
      self.startup()
      print('\nThe input file will be validated against the documented syntax, no')
      print('files will be created or modified. Note that there is no guarantee that')
      print('a validated input will actually run as intended, but a validation error')
      print('will most likely result in a run error.')
    self.rc = 0
    # flatten input to keep only programs (recursively inside groups)
    flat = self.flow.contents[:]
    redo = True
    while (redo):
      redo = False
      new = []
      for item in flat:
        if (isinstance(item, abstract_flow.Program)):
          new.append(item)
        else:
          redo = True
          if (isinstance(item, abstract_flow.Group)):
            new.extend(item.contents)
      flat = new
    final_rc = 0
    for item in (flat):
      inp = '&' + '\n'.join(item.lines[1:])
      inputlines = sanitize(inp).strip().split('\n')
      inputlines.append('End of input')
      rc, result = validate(inputlines, self.keywords)
      print('\n--')
      for line in result:
        print(line)
      if (rc == 0):
        print('\nLooks good!')
      elif (rc > 0):
        print('\n*** FAILED! ***')
      print('--')
      final_rc = max(final_rc, rc)
    if (final_rc):
      print('\n*************************************')
      print('*** There were validation errors. ***')
      print('*************************************')
    else:
      print('\nValidation passed.')
    self.rc = final_rc

  def auto(self):
    self.startup()
    set_utf8('MOLCAS_ITER', '0')
    if (get_utf8('MOLCAS_VALIDATE') == 'FIRST'):
      self.validate()
      if (self.rc > 0):
        self.rc = '_RC_INPUT_ERROR_'
        self.end()
        return
    try:
      self.run_logue('prologue')
      self.rc = self.flow.run(self)
      if (self.rc is None):
        self.rc = '_RC_ALL_IS_WELL_'
      if (self._goto):
        print('Label "{0}" not found'.format(self._goto))
        self.rc = '_RC_INPUT_ERROR_'
      if ((self.rc_to_name(self.rc) == '_RC_ALL_IS_WELL_') and (not self._last_module)):
        print('*** Extra labels in check file')
        self.rc = '_RC_CHECK_ERROR_'
      self.run_logue('epilogue')
    except KeyboardInterrupt:
      self.rc = '_RC_JOB_KILLED_'
    finally:
      self.end()

  @property
  def is_serial(self):
    serial = get_utf8('MOLCAS_SERIAL')
    nprocs = int(get_utf8('MOLCAS_NPROCS', default='1'))
    return ((serial == 'YES') or (nprocs == 1))

  def wrap_command(self, executable, may_be_serial=True, may_debug=True):
    '''
    Wrap the specified command as appropriate.
    If may_be_serial=False, never uses RUNBINARYSER.
    Returns a list.
    '''
    nprocs = int(get_utf8('MOLCAS_NPROCS', default='1'))
    set_utf8('MOLCAS_NPROCS', nprocs)
    if (may_debug):
      debugger = get_utf8('MOLCAS_DEBUGGER')
    else:
      debugger = ''
    if (debugger):
      command = '{0} $program'.format(debugger)
    elif (may_be_serial and self.is_serial):
      command = self._rte['RUNBINARYSER']
    else:
      command = self._rte['RUNBINARY']
    command = sub(r'\$program', executable, command)
    command = shsplit(expandvars(command, default='UNKNOWN_VARIABLE'))
    return command

  def _set_threads(self):
    threads = get_utf8('MOLCAS_THREADS')
    if (threads is not None):
      try:
        threads = int(threads)
        if (threads == 0):
          if ('OMP_NUM_THREADS' in environ):
            del environ['OMP_NUM_THREADS']
        else:
          set_utf8('OMP_NUM_THREADS', threads)
      except:
        pass

  def run_module(self, name, inp):
    if (self._ready):
      set_utf8('MOLCAS_CURRENT_PROGRAM', name.lower())
      self._set_threads()
      self.write_environment()
      try:
        self._current_module = Molcas_module(self, name, inp)
      except:
        self.rc = '_RC_NOT_AVAILABLE_'
        return self.rc
      rc = self._current_module.run()
      rcname = ''
      if (rc is not None):
        self.rc = rc
        rcname = self.rc_to_name(self.rc)
      if (rcname == '_RC_INVOKED_OTHER_MODULE_'):
        self._nest_level += 1
        set_utf8('EMIL_RC2',self._nest_level)
        stdin_name = 'Stdin.{0}.{1}'.format(self._nest_level, self._loop_level)
        Stdin = join(self.scratch, stdin_name)
        try:
          new_inp, files = EMIL_Parse(Stdin)
        except EMILException as message:
          raise MolcasException('EMIL: {0}'.format(message))
        self.add_files(files)
        rc = new_inp.run(self)
        if (rc is None):
          self.rc = '_RC_ALL_IS_WELL_'
        else:
          self.rc = rc
          rcname = self.rc_to_name(self.rc)
        self._nest_level -= 1
        set_utf8('EMIL_RC2',self._nest_level)
      # I wouldn't put this here, but compatibility obliges
      if ((rcname in ['_RC_ALL_IS_WELL_', '_RC_CONTINUE_LOOP_', '_RC_CONTINUE_UNIX_LOOP_', '_RC_EXIT_EXPECTED_', '_RC_INVOKED_OTHER_MODULE_']) or \
          (get_utf8('MOLCAS_TRAP', default='on').lower() == 'off')):
        self._run_check()
    else:
      print('***************************************************')
      print('*** MOLCAS system not ready -- setup not called ***')
      print('***************************************************')
      self.rc = '_RC_GENERAL_ERROR_'
    return self.rc

  def _run_check(self):
    '''
    Run the check command for verification.
    Give preference to the compiled "CHECK" program, but use a python method if not available.
    '''
    check = get_utf8('MOLCAS_TEST', default='').lower()
    if ((check == 'check') or (check=='gene')):
      set_utf8('MOLCAS_CURRENT_PROGRAM', 'check')
      self._check_count += 1
      self.write_environment()
      check_input = 'check {0}'.format(self._check_count)
      try:
        self._current_module = Molcas_module(self, 'check', check_input)
        rc = self._current_module.run()
        if (rc == '_RC_NOT_AVAILABLE_'):
          raise Exception
      except:
        (rc, last) = check_test(join(self.scratch, 'molcas_info'), join(self.scratch, 'checkfile'), self._check_count)
        self._last_module = last
      if (self.rc_to_name(rc) != '_RC_ALL_IS_WELL_'):
        self.rc = rc
    self.parallel_task(['x', 'molcas_info'], force=True)

  #TODO: buffer parnell calls
  def parallel_task(self, task, force=False):
    task_type = task[0][0]
    # some replacements for parnell calls in the serial case,
    # if no replacement is available, parnell will be used
    if (self.is_serial):
      # copy files: should also work if the destination is a dir
      if (task_type == 'c'):
        orig = task[2]
        if (not isabs(orig)):
          orig = join(self.scratch, orig)
        dest = task[3]
        if (not isabs(dest)):
          dest = join(self.scratch, dest)
        try:
          copy2(orig, dest)
        # would use SameFileError, but that's only available since python 3.4,
        # so use this workaround
        except Error as e:
          if ('same file' not in text_type(e)):
            raise
        except FileNotFoundError as e:
          if (e.errno == ENOENT):
            if (not force):
              print('Error: file "{0}" not found'.format(orig))
            return -1
          else:
            raise
        return 0
      # remove files: the list is colon-separated, and based on the scratch dir
      if (task_type == 'x'):
        rc = 0
        for i in task[1].split(':'):
          try:
            orig = join(self.scratch, i)
            remove(orig)
          except FileNotFoundError as e:
            if (e.errno == ENOENT):
              if (not force):
                print('Error: file "{0}" not found'.format(orig))
              rc -= 1
            else:
              raise
        return rc
    output = BytesIO()
    error = BytesIO()
    if (task_type == 'b'):
      # No working directory for "base"
      cwd = None
    else:
      cwd = self.scratch
    parnell = self.wrap_command(self.parnell, may_be_serial=False, may_debug=False)
    rc = teed_call(parnell + task, cwd=cwd, stdout=output, stderr=error)
    return rc

  def delete_scratch(self, force=False):
    #TODO: use parnell
    if (self._ready or force):
      if (realpath(self.scratch) == realpath(self.currdir)):
        line = '*** WorkDir and CurrDir are the same, not cleaned! ***'
      else:
        # WorkDir may be a link and not a real directory, so remove its contents
        try:
          filelist = listdir(self.scratch)
        except:
          filelist = []
        for i in [join(self.scratch, x) for x in filelist]:
          if (isdir(i)):
            rmtree(i)
          else:
            remove(i)
        line = '*** WorkDir at {0} cleaned ***'.format(self.scratch)
      print('*'*len(line))
      print(line)
      print('*'*len(line))

  def in_sbin(self, prog):
    '''Return the path of a program in sbin if it exists'''
    if (prog == basename(prog)):
      for path in [self.molcas] + self.sources:
        filename = join(path, 'sbin', prog)
        if (isfile(filename) and access(filename, X_OK)):
          return filename
    return False

  def run_sbin(self, command):
    '''Run a program from the sbin directory if it exists'''
    filename = self.in_sbin(command[0])
    if (filename):
      command[0] = filename
      rc = teed_call(command)
      return rc
    else:
      return None

  def run_logue(self, logue_name):
    '''Run a prologue/epilogue, if found'''
    files = []
    logue_dir = get_utf8('MOLCAS_LOGUE_DIR')
    if (logue_dir):
      if (not isabs(logue_dir)):
        logue_dir = join(self.currdir, logue_dir)
      files.append(join(logue_dir, logue_name))
    files.append(dotmolcas(logue_name))
    for filename in files:
      if (isfile(filename) and access(filename, X_OK)):
        # No rc control here, if it fails, so be it
        output = BytesIO()
        error = BytesIO()
        teed_call(['sh', filename], cwd=self.scratch, stdout=output, stderr=error)

  def check_license(self):
    '''Use molcas.exe to check the license'''
    filename = join(self.molcas, 'bin', 'molcas.exe')
    if (isfile(filename) and access(filename, X_OK)):
      try:
        out = check_output(filename, stderr=STDOUT).decode('utf-8')
        rc = 0
      except CalledProcessError as e:
        out = e.output.decode('utf-8')
        rc = e.returncode-1
      # Capture the licensee
      match = search(r'This copy of MOLCAS is licensed to\s*(.*)\n', out)
      if (match):
        self.licensee = match.group(1)
      return rc
    else:
      return None

#===============================================================================

_prgm_match = re_compile(r'\s*\(prgm\)\s+(\S+)\s+(executable|script)')
_file_match = re_compile(r'\s*\(file\)\s+(\S+)\s+(\S+)\s+(\S+)')
_rc_match = re_compile(r'\$(\w+)\s*=\s*(\d+)\s*;')
_emil_newln = re_compile(r'[ \t]*[=][ \t]*')
_emil_leadb = re_compile(r'^[ \t]*', flags=MULTILINE)
_emil_endofinput = re_compile(r'^end\s*of\s*input', flags=MULTILINE|IGNORECASE)

# TODO: custom .prgm, gracefully fail
def parse_prgm(prgm_file):
  '''Parse a .prgm file to find the module executable and files.'''
  prgm = None
  file_list = {}
  with utf8_open(prgm_file, 'r') as p_f:
    for line in p_f:
      match = _prgm_match.match(line)
      if (match):
        prgm = [expandvars(match.group(1), default='UNKNOWN_VARIABLE').replace('"', ''), match.group(2)[0]]
      match = _file_match.match(line)
      if (match):
        file_list[match.group(1)] = [expandvars(match.group(2), default='UNKNOWN_VARIABLE').replace('"', ''), match.group(3)]
  return (prgm, file_list)

def sanitize(text):
  '''Sanitize input, breaking lines at '=', removing leading blanks, etc'''
  out_text = _emil_leadb.sub('', text)
  out_text = _emil_newln.sub('\n', out_text)
  out_text = _emil_endofinput.sub('End of input', out_text)
  return out_text

class Molcas_module(object):
  '''Setup and run a Molcas module'''

  # Module initialization
  # 1st argument: Molcas environment object
  # 2nd argument: module name
  # 3rd argument (optional): input
  def __init__(self, parent, *args):
    if (len(args) < 1):
      raise MolcasException('Bad initialization of module')
    self.parent = parent
    if (not isinstance(self.parent, Molcas_wrapper)):
      raise MolcasException('The module is not contained in a Molcas_wrapper instance')
    name = args[0].lower()
    if name in self.parent.alias:
      self.name = self.parent.alias[name]
    else:
      self.name = name
    # Read the .prgm file to find the executable and the list of module files
    try:
      prgm_file = join(self.parent.molcas, 'data', self.name + '.prgm')
      (self._exec, self._files) = parse_prgm(prgm_file)
    except FileNotFoundError as e:
      if (e.errno == ENOENT):
        raise MolcasException('Unknown module: {0}'.format(self.name))
      else:
        raise
    prgm_file = join(self.parent.molcas, 'data', 'global.prgm')
    self._files.update(parse_prgm(prgm_file)[1])
    self._links = []
    self.rc = 0
    # Pass the input
    if (len(args) >= 2):
      inp = args[1]
      self.module_input = inp

  # Define the module_input property
  @property
  def module_input(self):
    return self._input
  @module_input.setter
  def module_input(self, inp):
    self._input = sanitize(inp)

  # Get the start line with time
  @property
  def start(self):
    return '--- Start Module: {0} at {1} ---'.format(self.name, self._start.ctime())

  # Get the end line with time and return code
  @property
  def stop(self):
    return '--- Stop Module: {0} at {1} /rc={2} ---'.format(self.name, self._stop.ctime(), self.rc_name)

  # Get the line with time spent (empty if less than 1 second)
  @property
  def span(self):
    span = self._stop - self._start
    d = span.days
    h = span.seconds//3600
    m = (span.seconds-h*3600)//60
    s = span.seconds-h*3600-m*60
    d_name = 'day' if (d==1) else 'days'
    h_name = 'hour' if (h==1) else 'hours'
    m_name = 'minute' if (m==1) else 'minutes'
    s_name = 'second' if (s==1) else 'seconds'
    if (d > 0):
      string = '--- Module {{0}} spent {{1}} {0} {{2}} {1} {{3}} {2} {{4}} {3} ---'.format(d_name, h_name, m_name, s_name)
    elif (h > 0):
      string = '--- Module {{0}} spent {{2}} {1} {{3}} {2} {{4}} {3} ---'.format(d_name, h_name, m_name, s_name)
    elif (m > 0):
      string = '--- Module {{0}} spent {{3}} {2} {{4}} {3} ---'.format(d_name, h_name, m_name, s_name)
    elif (s > 0):
      string = '--- Module {{0}} spent {{4}} {3} ---'.format(d_name, h_name, m_name, s_name)
    else:
      string = ''
    return string.format(self.name, d, h, m, s)

  # Get a tuple with wall/user/system timing
  @property
  def timing(self):
    wall_time = (self._stop - self._start).total_seconds()
    user_time = self._rstop.ru_utime - self._rstart.ru_utime
    system_time = self._rstop.ru_stime - self._rstart.ru_stime
    return (wall_time, user_time, system_time)

  # Get the output stream
  @property
  def output(self):
    return (bytes(self.start, 'utf-8') + b'\n' + self._output.getvalue() + b'\n' + bytes(self.stop, 'utf-8'))

  # Get the error stream
  @property
  def error(self):
    return (bytes(self.start, 'utf-8') + b'\n' + self._error.getvalue() + b'\n' + bytes(self.stop, 'utf-8'))

  # Get the named return code (if available)
  @property
  def rc_name(self):
    return self.parent.rc_to_name(self.rc)

  # Private method to write the input in the $WorkDir
  def _write_input(self, inp):
    stdin = join(self.parent.scratch, 'stdin')
    with utf8_open(stdin, 'w') as i_f:
      print('&' + self.name.upper(), file=i_f)
      if (inp):
        print(inp.strip(), file=i_f)
      print('End of input', file=i_f)
    input_file = self.name.upper()[:5] + 'INP'
    if (input_file in self._files):
      self.parent.parallel_task(['c', '1', stdin, self._files[input_file][0]], force=True)
    self.parent.parallel_task(['c', '1', stdin, self.parent.scratch], force=True)

  def _validate_input(self, inp):
    val = get_utf8('MOLCAS_VALIDATE', default='NO').upper()
    if (val not in ['YES', 'CHECK']):
      return
    if (self.name == 'check'):
      return
    rc, result = validate(inp, self.parent.keywords)
    print('\n-- Input validation for {}'.format(self.name))
    for line in result:
      print(line)
    if (rc == 0):
      print('\nLooks good!')
    elif (rc > 0):
      print('\n*** FAILED! ***')
    print('--')
    if (val == 'YES'):
      self.rc = rc

  # Private method to read the return code from the $WorkDir
  def _read_rc(self):
    rc_local = join(self.parent.scratch, 'rc.local')
    rc_global = join(self.parent.scratch, 'rc.global')
    self.parent.parallel_task(['c', '4', rc_local, rc_global], force=True)
    with utf8_open(rc_global, 'r') as rc_f:
      self.rc = int(rc_f.read())

  def _read_extra_prgm(self):
    prgm_file = join(self.parent.scratch, 'extra.prgm')
    if (exists(prgm_file)):
      self._files.update(parse_prgm(prgm_file)[1])

  def _copy_files(self):
    self._read_extra_prgm();
    dest = self.parent.output
    if (dest == self.parent.scratch):
      return
    files_to_copy = sorted([(k,v[0]) for (k,v) in self._files.items() if 's' in v[1]])
    files_to_move = sorted([(k,v[0]) for (k,v) in self._files.items() if 'm' in v[1]])
    files = self._copy_or_move(copy2, dest, files_to_copy)
    files.extend(self._copy_or_move(move, dest, files_to_move))
    if (len(files) > 0):
      listfiles = ' '.join(files)
      print(fill(listfiles, width=180, break_long_words=False, break_on_hyphens=False,
               initial_indent='*** files: ',
            subsequent_indent='           '))
      print('    saved to directory {0}'.format(dest))

  def _copy_or_move(self, action, dest, filelist):
    #TODO: use parnell
    #TODO: support the '.' attribute
    files = []
    for name, path in filelist:
      if '*' in self._files[name][1]:
        path += '*'
      for i in glob(path):
        if (datetime.fromtimestamp(getmtime(i)) > self._start):
          bi = basename(i)
          j = join(dest, bi)
          if (exists(j)):
            if (isfile(j)):
              if (self.parent.save_mode == 'repl'):
                action(i, j)
              elif (self.parent.save_mode == 'orig'):
                orig = j+'.orig'
                if (not exists(orig)):
                  move(j, orig)
                action(i, j)
              elif (self.parent.save_mode == 'incr'):
                fmt = '{0}.#{1}#'
                n = 1
                jj = fmt.format(j, n)
                while (exists(jj)):
                  n += 1
                  jj = fmt.format(j, n)
                action(i, jj)
              files.append(bi)
            else:
              pass
          else:
            action(i, j)
            files.append(bi)
    return files

  def _delete_files(self):
    files_to_delete = [(k,v[0]) for (k,v) in self._files.items() if 'p' in v[1]]
    rmlist = ['extra.prgm']
    for name, path in files_to_delete:
      if '*' in self._files[name][1]:
        path += '*'
      for i in glob(path):
        rmlist.append(relpath(i, self.parent.scratch))
    for i in glob(join(self.parent.scratch, 'purge*')):
      rmlist.append(relpath(i, self.parent.scratch))
    for i in self._links:
      rmlist.append(relpath(i, self.parent.scratch))
    if (len(rmlist) > 0):
      self.parent.parallel_task(['x', ':'.join(rmlist)], force=True)

  def _make_links(self):
    if (self.name == 'check'):
      return
    if (self.name == 'scf'):
      orblist = ['GVORB', 'LOCORB', 'UHFORB', 'SCFORB', 'RASORB', 'PT2ORB', 'UNAORB', 'CIORB01', 'CPFORB', 'SIORB', 'GSSORB']
    elif (self.name == 'grid_it'):
      orblist = ['GVORB', 'LOCORB', 'SIORB', 'PT2ORB', 'CIORB01', 'CPFORB', 'RASORB', 'MP2ORB', 'UHFORB', 'UNAORB', 'SCFORB', 'GSSORB']
    else:
      orblist = ['GVORB', 'LOCORB', 'RASORB', 'PT2ORB', 'UNAORB', 'SCFORB', 'CIORB01', 'CPFORB', 'SIORB', 'GSSORB']
    #TODO: fail gracefully if file does not exist
    prgm_file = join(self.parent.molcas, 'data', 'inporb_files.prgm')
    filelist = parse_prgm(prgm_file)[1]
    orblist = [filelist[x][0] for x in orblist if x in filelist]
    print('')
    inporb = join(self.parent.scratch, 'INPORB')
    if (not exists(inporb)):
      for orbfile in orblist:
        if (exists(orbfile)):
          symlink(orbfile, inporb)
          self.parent.parallel_task(['c', '1', inporb, self.parent.scratch], force=True)
          self._links.append(inporb)
          print('*** symbolic link created: INPORB -> {0}'.format(basename(orbfile)))
          break

  # Run the module
  def run(self):
    if (self._exec is None):
      return '_RC_NOT_AVAILABLE_'

    self._write_input(self.module_input)

    if (self.name == 'gateway'):
      runfiles = ['RUNFILE']
      if ('RUNFILE' in self._files):
        runfiles.append(relpath(self._files['RUNFILE'][0], self.parent.scratch))
      self.parent.parallel_task(['x', ':'.join(runfiles)], force=True)
    elif (self.name == 'loop'):
      set_utf8('MOLCAS_REDUCE_PRT', 'NO')

    command = self.parent.wrap_command(self._exec[0])
    no_tee=(get_utf8('MOLCAS_DEBUGGER') is not None)

    self._output = BytesIO()
    self._error = BytesIO()
    self._start = datetime.now()
    self._make_links()
    print(self.start)
    self._rstart = getrusage(RUSAGE_CHILDREN)
    self._validate_input(join(self.parent.scratch, 'stdin'))
    if (self.rc > 0):
      self.rc = '_RC_INPUT_ERROR_'
    else:
      self.parent.run_logue('module.prologue')
      if (isfile(self._exec[0]) and access(self._exec[0], X_OK)):
        teed_call(command, cwd=self.parent.scratch, stdout=self._output, stderr=self._error, no_tee=no_tee)
        self._read_rc()
      else:
        self.rc = '_RC_NOT_AVAILABLE_'
      self.parent.run_logue('module.epilogue')
    self._rstop = getrusage(RUSAGE_CHILDREN)
    self._stop = datetime.now()
    print(self.stop)
    self._copy_files()
    self._delete_files()
    span = self.span
    if (span):
      print(span)

    self.parent.add_resources(self.timing)
    k = self.output

    return self.rc



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2019,2020, Ignacio Fdez. Galván                        *
#***********************************************************************

from __future__ import (unicode_literals, division, absolute_import, print_function)
import re
import shlex
try:
  from lxml import etree as ET
except ImportError:
  pass

# Global variables
class gv:
  current_name = None
  lookup = { 'NSYM': None }
  syms = [ 1, 2, 4, 8 ]

# Basic expression to find keywords:
#  KEYWORD elements
#  GROUP elements
#  SELECT/KEYWORD elements
#  all of the above inside GROUP[KIND=BOX] (just visual grouping)
box = '(.|{0}|{0}/{0})'.format('GROUP[@KIND="BOX"]')
kw_exp = '{0}/GROUP[@KIND!="BOX"] | {0}/KEYWORD | {0}/SELECT/KEYWORD'.format(box)

# Compare two strings, case-insensitive, trimmed/padded to specified length
# (by default, 4 characters, typical of Molcas keywords)
def cmp_str(string1, string2, n=4):
  s1 = re.sub(r'\s+', ' ', string1.strip().upper()).ljust(n)[0:n]
  s2 = re.sub(r'\s+', ' ', string2.strip().upper()).ljust(n)[0:n]
  return s1 == s2

# Convert an integer number from Fortran format
def fortran_int(num):
  # pretend environment variables are always good
  if (num[0] == '$'):
    return 1
  return int(num)

# Convert a real number from Fortran format
def fortran_float(num):
  # pretend environment variables are always good
  if (num[0] == '$'):
    return 1.0
  # in case there is no exponent marker
  num = re.sub(r'([\d.])([+-]\d)', r'\1e\2', num)
  # convert d/D to e
  num = re.sub(r'[dD]', 'e', num)
  return float(num)

# Convert a list to integers
def to_int(parts):
  return [fortran_int(i) for i in parts]

# Convert a list to floats
def to_float(parts):
  return [fortran_float(i) for i in parts]

# Split in Fortran style
def fortran_split(string):
  return string.replace(',', ' ').split()

# Split in Fortran style, with quotes
def fortran_split_quotes(string):
  return shlex.split(string.replace(',', ' '))

# Get the first word
def first_word(string):
  return string.split()[0].strip()

# Get the first word as an integer
def first_int(string):
  return fortran_int(first_word(string))

# Is line blank?
def blank(string):
  return string.strip() == ''

# Check if the current line starts a group
#  1: group matches, consume one line
#  0: group doesn't match
# -1: group matches, but unhandled type
def test_group(lines, group):
  name = group.get('NAME').upper().strip()
  kind = group.get('KIND')
  gv.current_name = name
  if (cmp_str(name, lines[0])):
    if (kind in ['BLOCK', 'RADIO']):
      return 1
    else:
      return -1
  return 0

# Check if the current line(s) matches a keyword
#    n: the next n lines match the keyword
#   -n: the next n lines match the keyword and end a group
# None: the keyword doesn't match
def test_keyword(lines, keyword):
  name = keyword.get('NAME').upper().strip()
  kind = keyword.get('KIND')
  size = keyword.get('SIZE')
  choice = keyword.get('LIST')
  gv.current_name = name
  # if the keyword is in a group, these are the ways to end the group
  endlist = ['END']
  parent = keyword.getparent()
  if (parent.tag == 'GROUP'):
    endlist.append('END' + parent.get('NAME')[0])
  l = 0
  names = [name]
  also = keyword.get('ALSO')
  if (also):
    names.extend(keyword.get('ALSO').split(','))
  if (any([cmp_str(i, lines[0]) for i in names])):
    ll = test(lines, keyword, kind, size, choice)
    # try possible <ALTERNATE> definitions
    if (ll is None):
      for alt in (keyword.xpath('ALTERNATE')):
        kind_ = alt.get('KIND')
        size_ = alt.get('SIZE')
        choice_ = alt.get('LIST')
        ll = test(lines, keyword, kind_, size_, choice_)
        if (ll is not None):
          l += ll
          break
      else:
        return None
    else:
      l += ll
    # if tag is in endlist, invert
    if (name in endlist):
      l *= -1
  else:
    return None
  return l

# Wrapper function to test a keyword with standard or custom kind
def test(lines, keyword, kind, size, choice=None):
  if (kind == 'CUSTOM'):
    return test_custom(lines, keyword)
  else:
    return test_standard(lines, kind, size, choice=choice)

# Check if the current lines match a keyword with standard types
#    n: the next n lines match the keyword
# None: the keyword doesn't match
def test_standard(lines, kind, size, computed=False, choice=None):

  # first line contains the keyword
  l = 1
  # no additional input
  if (kind == 'SINGLE'):
    pass

  # string types
  elif (kind == 'STRINGS'):
    if (size is None):
      return None
    n = int(size)
    if (computed):
      n *= first_int(lines[l])
      l += 1
    l += n
    if (l >= len(lines)):
      return None
  elif (kind == 'STRING'):
    return test_standard(lines, 'STRINGS', 1)
  elif (kind == 'STRINGS_COMPUTED'):
    return test_standard(lines, 'STRINGS', size, computed=True)
  elif (kind == 'STRINGS_LOOKUP'):
    if (gv.lookup.get(size) is not None):
      return test_standard(lines, 'STRINGS', gv.lookup[size])
    else:
      return None

  # int types
  elif (kind == 'INTS'):
    if (computed):
      off = 1
      n = 0
    else:
      off = 0
      n = int(size)
    ints = 0
    # read the necessary number of integers
    while (ints < n+off):
      if (l >= len(lines)):
        return None
      for part in fortran_split(lines[l]):
        try:
          i = fortran_int(part)
          ints += 1
          if ((gv.lookup.get(gv.current_name) is None) and (ints == 1)):
            gv.lookup[gv.current_name] = i
        except:
          return None
        if (computed and (n == 0)):
          n = i*int(size)
          off = 0
          ints = 0
        if (ints == n+off):
          break
      l += 1
  elif (kind == 'INT'):
    return test_standard(lines, 'INTS', 1)
  elif (kind == 'INTS_COMPUTED'):
    return test_standard(lines, 'INTS', size, computed=True)
  elif (kind == 'INTS_LOOKUP'):
    if (gv.lookup.get(size) is not None):
      return test_standard(lines, 'INTS', gv.lookup[size])
    elif (size == 'NSYM'):
      for n in reversed(gv.syms):
        ll = test_standard(lines, 'INTS', n)
        if (ll):
          l = ll
          gv.lookup[size] = n
          break
    elif (size == 'ANY'):
      n = 1
      ll = 0
      while (ll is not None):
        ll = test_standard(lines, 'INTS', n)
        if (ll):
          l = ll
          n += 1
        else:
          if (n == 1):
            return None

  # real types
  elif (kind == 'REAL'):
    return test_standard(lines, 'REALS', 1)
  elif (kind == 'REALS'):
    if (computed):
      off = 1
      n = 0
    else:
      off = 0
      n = int(size)
    nums = 0
    # read the necessary number of reals
    while (nums < n+off):
      if (l >= len(lines)):
        return None
      for part in fortran_split(lines[l]):
        try:
          if (off == 0):
            i = fortran_float(part)
          else:
            i = fortran_int(part)
          nums += 1
        except:
          return None
        if (computed and (n == 0)):
          n = i*int(size)
          off = 0
          nums = 0
        if (nums == n+off):
          break
      l += 1
  elif (kind == 'REALS_COMPUTED'):
    return test_standard(lines, 'REALS', size, computed=True)
  elif (kind == 'REALS_LOOKUP'):
    if (gv.lookup.get(size) is not None):
      return test_standard(lines, 'REALS', gv.lookup[size])
    elif (size == 'NSYM'):
      for n in reversed(gv.syms):
        ll = test_standard(lines, 'REALS', n)
        if (ll):
          l = ll
          gv.lookup[size] = n
          break
    elif (size == 'DEG_FREEDOM'):
      n = 1
      ll = 0
      while (ll is not None):
        ll = test_standard(lines, 'REALS', 3*n)
        if (ll):
          l = ll
          gv.lookup[size] = n
          n += 1

  # choice type
  elif (kind == 'CHOICE'):
    opts = [i.upper().strip() for i in choice.split(',')]
    opts = [i.split(':')[0].strip() for i in opts if (i != '----')]
    ll = 0
    for opt in opts:
      # an option can contain several lines, separated by ;
      words = [i.strip() for i in opt.split(';')]
      ll = 0
      for i,w in enumerate(words):
        word = first_word(lines[i+1])
        # pretend an environment variable will always match
        if ((word.strip()[0] != '$') and (word.strip().upper() != w)):
          ll = 0
          break
        ll += 1
      if (ll > 0):
        break
    if (ll == 0):
      return None
    l += ll

  # unrecognized type
  else:
    return None
  return l

# Check if the current lines match a keyword with custom type,
# these are keywords with a syntax that doesn't comply with the standard types
#    n: the next n lines match the keyword
# None: the keyword doesn't match
def test_custom(lines, keyword):

  name = keyword.get('NAME')
  module = keyword.get('MODULE')
  node = keyword
  # first line contains the keyword
  l = 1

  if (module == 'CASPT2'):
    if (name == 'AFREEZE'):
      try:
        parts = fortran_split(lines[l])
        n = fortran_int(parts[0])
        nums = to_float(parts[1:3])
        assert (len(nums) == 2)
        l += 1
        parts = fortran_split(lines[l])
        assert (len(parts) == n)
        l += 1
      except:
        return None
    elif (name == 'EFFE'):
      try:
        n = first_int(lines[l])
        ll = test_standard(lines, 'REALS', n*n)
        if (ll):
          l = ll
        else:
          return None
      except:
        return None
    elif (name in ['MULTISTATE', 'XMULTISTATE']):
      try:
        assert (first_word(lines[l]).upper() == 'ALL')
        l += 1
      except:
        return None
    else:
      return None

  elif (module == 'CASVB'):
    if (name == 'ORB'):
      try:
        l -= 1
        n = 0
        parts = fortran_split(lines[l])
        parts.pop(0)
        if (len(parts) < 1):
          l += 1
          parts = fortran_split(lines[l])
        n = fortran_int(parts.pop(0))
        if (len(parts) < 1):
          l += 1
          parts = fortran_split(lines[l])
        nums = to_float(parts)
        l += 1
        while (l < len(lines)):
          parts = fortran_split(lines[l])
          try:
            nums = to_float(parts)
            l += 1
          except:
            break
      except:
        return None
    elif (name == 'ORBREL'):
      try:
        parts = fortran_split(lines[l])
        assert (len(parts) >= 3)
        nums = to_int(parts[0:2])
        l += 1
      except:
        return None
    elif (name == 'FIXORB'):
      try:
        assert (first_word(lines[l]).upper() in ['ALL', 'NONE'])
        l += 1
      except:
        return None
    else:
      return None

  elif (module == 'CHT3'):
    if (name == 'TITLE'):
      kwlist = [kw.get('NAME') for kw in keyword.getparent().xpath(kw_exp)]
      ll = 1
      # all lines (max 10) until a keyword matches
      while ((l+ll < len(lines)) and (ll < 10)):
        if (any(cmp_str(lines[l+ll], kw) for kw in kwlist)):
          break
        ll += 1
      l += ll
    else:
      return None

  elif (module == 'DMRGSCF'):
    if (name == 'SOCCUPY'):
      ll = 1
      while (l+ll < len(lines)):
        try:
          parts = fortran_split(lines[l+ll])
          assert all([i in '0ud2' for i in parts])
          ll += 1
        except:
          break
      if (ll == 0):
        return None
      l += ll
    else:
      return None

  elif (module == 'ESPF'):
    if (name == 'EXTERNAL'):
      try:
        n = first_int(lines[l])
        l += 1
        if (n == 0):
          try:
            ll = 0
            while (l+ll < len(lines)):
              parts = fortran_split(lines[l+ll])
              i = fortran_int(parts[0])
              nums = to_float(parts[1:11])
              assert (len(nums) == 10)
              l += 1
          except:
            pass
          l += ll
        elif (n > 0):
          for ll in range(n):
            nums = to_float(fortran_split(lines[l])[0:4])
            l += 1
        else:
          return None
      except:
        return None
    else:
      return None

  elif (module == 'EXTF'):
    if (name == 'LINEAR'):
      try:
        n = first_int(lines[l])
        l += 1
        n = first_int(lines[l])
        l += 1
        n = fortran_float(first_word(lines[l]))
        l += 1
        n = first_int(lines[l])
        assert (n in [0, 1])
        l += 1
      except:
        return None
    else:
      return None

  elif (module == 'FFPT'):
    if (name in ['DIPO', 'QUAD', 'OCTU', 'EFLD', 'EFGR']):
      ll = 0
      while (l+ll < len(lines)):
        try:
          parts = fortran_split(lines[l+ll])
          i = fortran_float(parts[1])
          if (name in ['DIPO', 'EFLD']):
            assert (parts[0].upper() in ['X', 'Y', 'Z'])
          elif (name in ['QUAD', 'EFGR']):
            assert (parts[0].upper() in ['XX', 'XY', 'XZ', 'YY', 'YZ', 'ZZ'])
          elif (name in ['OCTU']):
            assert (parts[0].upper() in ['XXX', 'XXY', 'XXZ', 'XYY', 'XYZ', 'XZZ', 'YYY', 'YYZ', 'YZZ', 'ZZZ'])
          ll += 1
        except:
          break
      if (ll == 0):
        return None
      else:
        l += ll
      if (name != 'DIPO'):
        try:
          parts = fortran_split(lines[l].replace(',', ' '))
          assert (parts[0].upper() == 'ORIG')
          nums = to_float(parts[1:4])
          assert (len(nums) == 3)
          l += 1
        except:
          if (name in ['EFLD', 'EFGR']):
            return None
    elif (name == 'GLBL'):
      while (l < len(lines)):
        try:
          parts = fortran_split_quotes(lines[l])
          assert (len(parts[0]) == 8)
          i = fortran_int(parts[1])
          i = fortran_float(parts[2])
          l += 1
        except:
          break
    elif (name == 'SELE'):
      try:
        n = first_int(lines[l])
        l += 1
        for i in range(n):
          parts = fortran_split(lines[l])
          assert (parts[0].upper() in ['.TRUE.', '.FALSE.', 'T', 'F'])
          n1, n2 = to_int(parts[1:3])
          l += 1
        for i in range(n-1):
          parts = fortran_split(lines[l])
          assert all([ii.upper() in ['.TRUE.', '.FALSE.', 'T', 'F'] for ii in parts[0:i-1]])
          l += 1
        parts = fortran_split(lines[l])
        nums = to_float(parts[0:3])
        assert (len(nums) == 3)
        l += 1
      except:
        return None
    else:
      return None

  elif (module in ['GATEWAY', 'SEWARD']):
    if (name in ['COORD', 'XYZ']):
      try:
        assert (first_word(lines[l])[0] != '$')
        n = first_int(lines[l])
      except:
        n = 0
      l += 1
      if (n > 0):
        try:
          l += 1
          for ll in range(n):
            parts = fortran_split(lines[l])
            nums = to_float(parts[1:4])
            assert (len(nums) == 3)
            l += 1
        except:
          return None
    elif (name == 'BASIS (NATIVE)'):
      try:
        while (blank(lines[l])):
          l += 1
        opts = lines[l].split('/')
        l += 1
        end = False
        if ((len(opts) > 1) and (opts[1].strip().upper() == 'INLINE')):
          while (l < len(lines)):
            if (blank(lines[l])):
              l += 1
              continue
            parts = fortran_split(lines[l])
            if (parts[-1] == '/'):
              parts.pop(-1)
            try:
              nums = to_float(parts)
              l += 1
            except:
              break
        if ('FRAGMENT' in opts[0].strip().upper()):
          ncoor = 0
          nener = 0
          while (l < len(lines)):
            parts = fortran_split(lines[l])
            if (blank(lines[l])):
              pass
            elif (parts[0].upper() == 'LBASIS'):
              ll = test_standard(lines[l:], 'STRINGS_COMPUTED', 1)
              if (ll):
                l += ll-1
              else:
                return None
            elif (parts[0].upper() == 'RELCOORDS'):
              ll = test_standard(lines[l:], 'REALS_COMPUTED', 3)
              if (ll):
                ncoor = fortran_int(lines[l+1].split()[0])
                l += ll-1
              else:
                return None
            elif (parts[0].upper() == 'ENERGIES'):
              ll = test_standard(lines[l:], 'REALS_COMPUTED', 1)
              if (ll):
                nener = fortran_int(lines[l+1].split()[0])
                l += ll-1
              else:
                return None
            elif (parts[0].upper() == 'MOCOEFF'):
              ll = test_standard(lines[l:], 'REALS_COMPUTED', nener)
              if (ll):
                l += ll-1
              else:
                return None
            elif (parts[0].upper() == 'MULLIKEN'):
              ll = test_standard(lines[l:], 'REALS', ncoor)
              if (ll):
                l += ll-1
              else:
                return None
            else:
              break
            l += 1
        elif ('ECP' in opts[0].strip().upper()):
          while (l < len(lines)):
            parts = fortran_split(lines[l])
            if (blank(lines[l])):
              pass
            elif (parts[0].upper() in ['M1', 'M2']):
              ll = test_standard(lines[l:], 'REALS_COMPUTED', 2)
              if (ll):
                l += ll-1
              else:
                return None
            elif (parts[0].upper() == 'COREREP'):
              ll = test_standard(lines[l:], 'REALS', 1)
              if (ll):
                l += ll-1
              else:
                return None
            elif (parts[0].upper() == 'PROJOP'):
              ll = 1
              n = first_int(lines[l+ll])
              ll += 1
              for i in range(n+1):
                n1, n2 = to_int(fortran_split(lines[l+ll])[0:2])
                ll += test_standard(lines[l+ll:], 'REALS', n2+n1*(1+n2))
              if (ll):
                l += ll-1
            else:
              break
            l += 1
        n = 0
        while (l < len(lines)):
          if (cmp_str(lines[l], 'END')):
            l += 1
            break
          parts = fortran_split(lines[l])
          if (blank(lines[l])):
            pass
          elif any([cmp_str(parts[0], i) for i in ['CHARGE', 'NUCLEAR']]):
            ll = test_standard(lines[l:], 'REALS', 1)
            if (ll):
              l += ll-1
            else:
              return None
          elif any([cmp_str(parts[0], i) for i in ['SPHERICAL', 'CARTESIAN', 'CONTAMINANT']]):
            assert (len(parts) > 1)
          elif any([cmp_str(parts[0], i) for i in ['MUON', 'PSEUDO', 'FRAGMENT']]):
            pass
          elif (cmp_str(parts[0], 'SPEC')):
            l += 1
            while (l < len(lines)):
              if (cmp_str(lines[l], 'END')):
                l += 1
                break
              elif (cmp_str(lines[l], 'VALE')):
                pass
              elif (cmp_str(lines[l], 'EXCH')):
                pass
              elif (cmp_str(lines[l], '1STO')):
                l += 1
              else:
                return None
              l += 1
            else:
              return None
            l -= 1
          else:
            parts = fortran_split(lines[l])
            nums = to_float(parts[1:4])
            assert (len(nums) == 3)
            n += 1
          l += 1
        else:
          return None
        if (n == 0):
          return None
      except:
        return None
    elif (name in ['CONSTRAINTS', 'NGEXCLUDE']):
      try:
        first = True
        while (l < len(lines)):
          lab = first_word(lines[l])
          l += 1
          if ((lab.upper() == 'INVERT') and (name == 'NGEXCLUDE')):
            continue
          if (cmp_str(lab, 'VALU')):
            first = False
            continue
          if (cmp_str(lab, 'END')):
            break
          if (first):
            typ = first_word(lines[l]).upper()
            assert (typ in ['BOND', 'ANGLE', 'LANGLE(1)', 'LANGLE(2)', 'DIHEDRAL', 'OUTOFP', 'DISSOC', 'CARTESIAN', 'EDIFF', 'SPHERE', 'TRANSVERSE', 'FRAGMENT'])
            l += 1
        else:
          return None
      except:
        return None
    elif (name in ['EPOT', 'EFLD', 'FLDG']):
      try:
        n = first_int(lines[l])
        l += 1
        for i in range(n):
          parts = fortran_split(lines[l])
          try:
            m = fortran_float(parts[0])
          except ValueError:
            pass
          else:
            nums = to_float(parts[0:3])
            assert (len(nums) == 3)
          l += 1
      except:
        return None
    elif (name == 'ISOTOPES'):
      try:
        n = first_int(lines[l])
        l += 1
        for i in range(n):
          parts = fortran_split(lines[l])
          if ((len(parts) > 2) and (parts[2].upper() == 'DALTON')):
            m = fortran_int(parts[0])
            m = fortran_float(parts[1])
          else:
            nums = to_int(parts[0:2])
            assert (len(nums) == 2)
          l += 1
      except:
        return None
    elif (name == 'REACTION'):
      try:
        parts = fortran_split(lines[l])
        n = fortran_int(parts.pop(2))
        nums = to_float(parts[0:3])
        assert (len(nums) >= 2)
        l += 1
      except:
        return None
    elif (name == 'RELATIVISTIC'):
      try:
        kw = first_word(lines[l]).upper()
        assert (kw[0] == 'R')
        n = int(kw[1:3])
        assert (kw[3] in ['O', 'E', 'S', 'M', 'C'])
        l += 1
      except:
        return None
    elif (name == 'TITLE'):
      kwlist = [kw.get('NAME') for kw in keyword.getparent().xpath(kw_exp)]
      ll = 1
      # all lines (max 10) until a keyword matches
      while ((l+ll < len(lines)) and (ll < 10)):
        if (any(cmp_str(lines[l+ll], kw) for kw in kwlist)):
          break
        ll += 1
      l += ll
    elif (name == 'XBAS'):
      lab = first_word(lines[l])
      l += 1
      if ('.' in lab):
        while (l < len(lines)):
          l += 1
          if (cmp_str(lines[l-1].split()[0], 'END')):
            break
        else:
          return None
    elif (name == 'ZMAT'):
      n = 0
      while (l < len(lines)):
        parts = lines[l].split()
        l += 1
        if ((len(parts) == 0) or cmp_str(parts[0], 'END')):
          break
        try:
          for i in range(n):
            num = fortran_int(parts[2*i+1])
            num = fortran_float(parts[2*i+2])
        except:
          return None
        n = min(n+1, 3)
      else:
        return None
    elif (name == 'XFIELD'):
      nmul = [0, 1, 4, 10]
      npol = [0, 1, 6]
      try:
        parts = fortran_split(lines[l])
        n = fortran_int(parts[0])
        if ((len(parts) > 1) and (parts[1].upper() == 'ANGSTROM')):
          parts.pop(1)
        l += 1
        nord = 1
        np = 0
        nfrag = 0
        nread = 0
        if (len(parts) > 1):
          nums = to_int(parts[1:5])
          if (len(nums) > 0):
            nord = nums[0]
          if (len(nums) > 1):
            np = nums[1]
          if (len(nums) > 2):
            nfrag = nums[2]
          if (len(nums) > 3):
            nread = nums[3]
        nint = nfrag+nread
        nfloat = 3+nmul[nord+1]+npol[np]
        for i in range(n):
          parts = fortran_split(lines[l])
          nums = to_int(parts[0:nint])
          assert (len(nums) == nint)
          nums = to_float(parts[nint:nint+nfloat])
          assert (len(nums) == nfloat)
          l += 1
      except:
        return None
    else:
      return None

  elif (module == 'GUESSORB'):
    if (name == 'PRMO'):
      try:
        parts = fortran_split(lines[l])
        n = fortran_int(parts[0])
        assert (n in [1, 2, 3, 4])
        if (len(parts) > 1):
          n = fortran_float(parts[1])
        l += 1
      except:
        return None
    else:
      return None

  elif (module in ['GUGA', 'GUGADRT']):
    if (name == 'REFERENCE'):
      try:
        parts = fortran_split(lines[l])
        n1, n2 = to_int(parts[0:2])
        l += 1
        for i in range(n1):
          s = first_word(lines[l])
          assert ((len(s) >= n2) and all([c in '012' for c in s]))
          l += 1
      except:
        return None
    else:
      return None

  elif (module == 'LOCALISATION'):
    if (name in ['LOCN', 'LOCC']):
      try:
        parts = fortran_split(lines[l])
        n = fortran_float(parts[1])
        n = fortran_int(parts[0])
        l += 1
        parts = fortran_split(lines[l])
        assert (len(parts) >= n)
        l += 1
      except:
        return None
    else:
      return None

  elif (module == 'MBPT2'):
    if (name == 'LOVMP2'):
      try:
        parts = fortran_split(lines[l])
        n = fortran_float(parts[1])
        n = fortran_int(parts[0])
        l += 1
        parts = fortran_split(lines[l])
        assert (len(parts) >= n)
        l += 1
      except:
        return None
    else:
      return None

  elif (module == 'POLY_ANISO'):
    if (name in ['PAIR', 'ALIN', 'LIN9']):
      try:
        n = first_int(lines[l])
        l += 1
        if (name == 'PAIR'):
          m = 1
        elif (name == 'ALIN'):
          m = 3
        elif (name == 'LIN9'):
          m = 9
        for i in range(n):
          parts = fortran_split(lines[l])
          nums = to_int(parts[0:2])
          assert (len(nums) == 2)
          nums = to_float(parts[2:2+m])
          assert (len(nums) == m)
          l += 1
      except:
        return None
    elif (name == 'HEXP'):
      try:
        n = first_int(lines[l])
        ll = test_standard(lines[l-1:], 'REALS_COMPUTED', 1)
        if ((ll is None) or (ll < 1)):
          return None
        l += ll-1
        ll = test_standard(lines[l-1:], 'REALS_COMPUTED', n+1)
        if ((ll is None) or (ll < 1)):
          return None
        l += ll-1
      except:
        return None
    elif (name == 'NNEQ'):
      try:
        parts = fortran_split(lines[l])
        n = fortran_int(parts[0])
        gv.lookup['3NNEQ'] = 3*n
        parts = [i.upper().strip() for i in parts[1:3]]
        assert (len(parts) == 2)
        assert all([i in ['T', 'F'] for i in parts])
        more = (parts[0] == 'F')
        l += 1
        parts = fortran_split(lines[l])
        nums = to_int(parts[0:n])
        assert (len(nums) == n)
        l += 1
        parts = fortran_split(lines[l])
        nums = to_int(parts[0:n])
        assert (len(nums) == n)
        l += 1
        if (more):
          parts = fortran_split(lines[l])[0:n]
          assert (len(parts) == n)
          parts = [i.upper().strip() for i in parts]
          assert all([i in ['A', 'B'] for i in parts])
          l += 1
          for i in parts:
            if (i == 'B'):
              n = fortran_float(first_word(lines[l]))
              l += 1
      except:
        return None
    elif (name == 'SYMM'):
      try:
        m = 1
        while (l < len(lines)):
          try:
            n = first_int(lines[l])
          except:
            break
          assert (n == m)
          m += 1
          l += 1
          while (l < len(lines)):
            ll = test_standard(lines[l-1:], 'REALS', 9)
            if ((ll is None) or (ll < 1)):
              break
            l += ll-1
        if (m == 1):
          return None
      except:
        return None
    else:
      return None

  elif (module == 'MCLR'):
    if (name == 'THERMO'):
      try:
        n = first_int(lines[l])
        l += 1
        n = fortran_float(first_word(lines[l]))
        l += 1
        while (l < len(lines)):
          if (cmp_str(lines[l].split()[0], 'END')):
            l += 1
            break
          n = fortran_float(first_word(lines[l]))
          l += 1
        else:
          return None
      except:
        return None
    else:
      return None

  elif (module == 'QMSTAT'):
    if (name == 'DPARAMETERS'):
      while (l < len(lines)):
        try:
          parts = fortran_split(lines[l])
          nums = to_float(parts[0:2])
          assert (len(nums) == 2)
          l += 1
        except:
          break
    elif (name == 'DISPERSION'):
      try:
        for ll in range(2):
          parts = fortran_split(lines[l])
          nums = to_float(parts[0:4])
          assert (len(nums) == 4)
          l += 1
      except:
        return None
      while (l < len(lines)):
        try:
          parts = fortran_split(lines[l])
          nums = to_float(parts[0:4])
          assert (len(nums) == 4)
          l += 1
        except:
          break
    elif (name == 'EXTERNAL'):
      try:
        n = first_int(lines[l])
        l += 1
        for i in range(n):
          parts = fortran_split_quotes(lines[l])
          m = fortran_float(parts[0])
          assert (len(parts[1]) == 8)
          m = fortran_int(parts[2])
          l += 1
      except:
        return None
    else:
      return None

  elif (module == 'QUATER'):
    if (name in ['GEO1', 'GEO2']):
      try:
        n = first_int(lines[l])
        l += 2
        for i in range(n):
          parts = fortran_split(lines[l])
          nums = to_float(parts[1:4])
          assert (len(nums) == 3)
          l += 1
        assert cmp_str(lines[l], 'END')
        l += 1
      except:
        return None
    else:
      return None

  elif (module == 'RASSCF'):
    if (name == 'CIROOT'):
      try:
        parts = fortran_split(lines[l])
        n1, n2 = to_int(parts[0:2])
        if (len(parts) > 2):
          n3 = fortran_int(parts[2])
        else:
          n3 = 0
        ll = 1
        if (n3 != 1):
          ll += test_standard(lines[l+ll-1:], 'INTS', n1) - 1
          if (n1 > 1):
            ll += test_standard(lines[l+ll-1:], 'REALS', n1) - 1
        l += ll
      except:
        return None
    elif (name == 'GASSCF'):
      try:
        parts = fortran_split(lines[l])
        n = fortran_int(parts[0])
        ll = 1
        for i in range(n):
          ll += test_standard(lines[l+ll-1:], 'INTS_LOOKUP', 'NSYM') - 1
          ll += test_standard(lines[l+ll-1:], 'INTS', 2) - 1
        l += ll
      except:
        return None
    elif (name == 'SUPSYM'):
      try:
        nsym = gv.lookup['NSYM']
        if (nsym is None):
          nsym = gv.syms[-1]
          find = True
        else:
          find = False
        for s in range(nsym):
          try:
            n = first_int(lines[l])
            l += 1
            for i in range(n):
              l += test_standard(lines[l-1:], 'INTS_COMPUTED', 1) - 1
          except:
            if (find and (s in gv.syms)):
              gv.lookup['NSYM'] = s
              break
            else:
              raise
      except:
        return None
    else:
      return None

  elif (module == 'RASSI'):
    if (name == 'NROFJOBIPHS'):
      try:
        parts = fortran_split(lines[l])
        n = fortran_int(parts[0])
        gv.lookup['NROFJOBIPHS'] = n
        if ((len(parts) > 1) and (parts[1].upper().strip() == 'ALL')):
          ll = 1
        else:
          ll = test_standard(lines, 'INTS_COMPUTED', 1) - l
          parts = fortran_split(' '.join(lines[l:l+ll]))
          nums = to_int(parts[1:])
          for i in range(n):
            ll += test_standard(lines[l+ll-1:], 'INTS', nums[i]) - 1
        l += ll
      except:
        return None
    elif (name == 'PROPERTY'):
      try:
        parts = fortran_split_quotes(lines[l])
        n = fortran_int(parts.pop(0))
        l += 1
        for i in range(n):
          if (len(parts) < 2):
            parts = fortran_split_quotes(lines[l])
            l += 1
          parts.pop(0)
          m = fortran_int(parts.pop(0))
      except:
        return None
    else:
      return None

  elif (module == 'SCF'):
    if (name == 'CONSTRAINTS'):
      try:
        parts = fortran_split(lines[l])
        n = to_int(parts)
        assert (len(n) in gv.syms)
        l += 1
        for m in n:
          parts = fortran_split(lines[l])
          nums = to_int(parts[0:m])
          assert (len(nums) == m)
          assert all([i in [-1, 1] for i in nums])
          l += 1
      except:
        return None
    elif (name == 'OCCUPIED'):
      ll = test_standard(lines[l-1:], 'INTS_LOOKUP', 'NSYM')
      if (ll is not None):
        l += ll-1
        ll = test_standard(lines[l-1:], 'INTS_LOOKUP', 'NSYM')
        if (ll is not None):
          l += ll-1
      else:
        return None
    elif (name in ['OCCNUMBERS', 'MCCNUMBERS']):
      ll = 0
      while (l+ll < len(lines)):
        parts = fortran_split(lines[l+ll])
        try:
          nums = to_float(parts)
        except:
          break
        ll += 1
      l += ll
    else:
      return None

  elif (module == 'SINGLE_ANISO'):
    if (name == 'QUAX'):
      try:
        n = first_int(lines[l])
        l += 1
        if (n == 3):
          for i in range(3):
            parts = fortran_split(lines[l])
            nums = to_float(parts[0:3])
            assert (len(nums) == 3)
            l += 1
      except:
        return None
    else:
      return None

  elif (module == 'SLAPAF'):
    if (name == 'RTRN'):
      try:
        parts = fortran_split(lines[l])
        n = fortran_int(parts[0])
        n = fortran_float(parts[1])
      except:
        return None
      l += 1
    elif (name in ['INTERNAL', 'TSCONSTRAINTS']):
      try:
        first = True
        while (l < len(lines)):
          lab = first_word(lines[l])
          l += 1
          if (cmp_str(lab, 'VARY')):
            first = False
            continue
          if (cmp_str(lab, 'FIX', 3)):
            first = False
            continue
          if (cmp_str(lab, 'VALU')):
            first = False
            continue
          if (cmp_str(lab, 'END')):
            break
          if (first):
            typ = first_word(lines[l]).upper()
            assert (typ in ['BOND', 'ANGLE', 'LANGLE(1)', 'LANGLE(2)', 'DIHEDRAL', 'OUTOFP', 'DISSOC', 'CARTESIAN', 'EDIFF', 'SPHERE', 'TRANSVERSE', 'FRAGMENT'])
            l += 1
        else:
          return None
      except:
        return None
    elif (name == 'THERMO'):
      try:
        n = first_int(lines[l])
        l += 1
        n = fortran_float(first_word(lines[l]))
        l += 1
        while (l < len(lines)):
          if (cmp_str(lines[l].split()[0], 'END')):
            l += 1
            break
          n = fortran_float(first_word(lines[l]))
          l += 1
        else:
          return None
      except:
        return None
    else:
      return None

  elif (module == 'VIBROT'):
    if (name in ['POTE', 'OBSE']):
      if (name == 'OBSE'):
        l += 1
        if (l >= len(lines)):
          return None
      ll = 0
      while (l+ll < len(lines)):
        try:
          parts = fortran_split(lines[l+ll])
          nums = to_float(parts[0:2])
          assert (len(nums) == 2)
          ll += 1
        except:
          if ((name != 'OBSE') and (ll == 0)):
            return None
          l += ll
          try:
            assert cmp_str(lines[l], 'PLOT')
            l += 1
            parts = fortran_split(lines[l])
            nums = to_float(parts[0:3])
            assert (len(nums) == 3)
            l += 1
            break
          except:
            break
    elif (name == 'ATOMS'):
      try:
        parts = fortran_split(lines[l])
        l += 1
        n = fortran_int(parts[0])
        if (n < 0):
          l += 1
        n = fortran_int(parts[2])
        if (n < 0):
          l += 1
      except:
        return None
    else:
      return None

  elif (module == 'WFA'):
    if (name == 'ATLISTS'):
      try:
        n = first_int(lines[l])
        l += 1
        for i in range(n):
          parts = lines[l].split()
          assert (parts.pop() == '*')
          nums = to_int(parts)
          l += 1
      except:
        return None
    elif (name == 'PROPLIST'):
      try:
        parts = lines[l].split()
        assert (parts[-1] == '*')
        l += 1
      except:
        return None
    else:
      return None
  else:
    return None
  return l

# Read and parse an XML keyword database from a file
def read_db(filename):
  root = ET.parse(filename).getroot()

  garbage = ET.Element('garbage')

  # Copy all content from INCLUDE references
  for include in root.xpath('//INCLUDE'):
    xmod = root.find('MODULE[@NAME="{0}"]'.format(include.get('MODULE')))

    this = include
    ignore = include.get('EXCEPT', '').split(',')
    for node in xmod:
      if (node.get('NAME') in ignore):
        continue
      new = ET.fromstring(ET.tostring(node))
      for i in new.xpath('//*'):
        if (i.get('NAME') in ignore):
          garbage.append(i)
      this.addnext(new)
      this = new
    garbage.append(include)

  # Add END keywords in modules and block groups if not already there
  for group in root.xpath('MODULE | //GROUP[@KIND="BLOCK"]'):
    if (group.find('KEYWORD[@NAME="END"]') is None):
      group.append(ET.Element('KEYWORD', NAME='END', KIND='SINGLE'))
    # groups can also end with ENDX (X is first letter in group's name)
    if (group.tag == 'GROUP'):
      end = 'END' + group.get('NAME')[0]
      if (group.find('KEYWORD[@NAME="{}"]'.format(end)) is None):
        group.append(ET.Element('KEYWORD', NAME=end, KIND='SINGLE'))

  del garbage

  return root

# This is a hack for the XYZ input of GATEWAY/SEWARD
# to hide "native input" keywords and enable "XYZ input" ones
def enable_xyz(module):
  for kw in module.xpath('//KEYWORD[@EXCLUSIVE="COORD"]'):
    kw.set('NAME', '#{}'.format(kw.get('NAME')))
  for kw in module.xpath('//KEYWORD[starts-with(@NAME, "*")]'):
    kw.set('NAME', kw.get('NAME')[1:])

# Validate a list of strings against a keyword database
def validate(inp, db):
  rc = 0
  result = []
  found = []

  # Re-initialize global variables
  gv.current_name = None
  gv.lookup = { 'NSYM': None }
  gv.syms = [ 1, 2, 4, 8 ]

  if (isinstance(inp, list)):
    inputlines = inp
  else:
    try:
      with open(inp, 'r') as f:
        inputlines = f.readlines()
    except:
      rc = -1
      return (rc, ['Input not found, no validation possible'])

  program = inputlines[0].strip().upper()
  if (program.startswith('&')):
    program = program[1:]
  else:
    rc = 2
    return (rc, ['No module name'])
  if (db is None):
    rc = -1
    return (rc, ['Database not found, no validation possible'])
  if (db == 'no_lxml'):
    rc = -2
    return (rc, ['lxml python module not found, no validation possible'])
  module = db.find('MODULE[@NAME="{0}"]'.format(program))
  if (module is None):
    rc = 3
    return (rc, ['No module named "{0}"'.format(program)])

  # This is a hack for the XYZ input of GATEWAY/SEWARD
  # to hide "XYZ input" keywords (they will be enabled if COORD, GROMACS or TINKER is found)
  if (module.get('NAME') in ['GATEWAY', 'SEWARD']):
    for kw in module.xpath('//KEYWORD[@NAME="BASIS (XYZ)"] | //KEYWORD[@NAME="GROUP"]'):
      kw.set('NAME', '*{}'.format(kw.get('NAME')))

  result.append('Input for: {0}'.format(program))

  stack = [None]
  group = module

  nline = 1
  bad = False
  # check line by line until the end of the file or the module input
  while (nline < len(inputlines) and group is not None):
    # skip blank lines (between keywords)
    if (blank(inputlines[nline])):
      nline += 1
      continue
    # check all allowed keywords in the current module/group
    anykw = False
    for kw in group.xpath(kw_exp):
      name = kw.get('NAME')
      # skip GUI-only keywords
      if (kw.get('LEVEL') == 'GUI'):
        continue
      if (kw.tag == 'KEYWORD'):
        # n should return the number of lines (if any) matching the keyword
        n = test_keyword(inputlines[nline:], kw)
        # for radio groups, the first matching keyword ends the group
        if ((group.get('KIND') == 'RADIO') and (n is not None)):
          n *= -1
        anykw = anykw or cmp_str(name, inputlines[nline])
        # if there's a match, print the line(s)
        if (n):
          if ((n < 0) and (len(stack) > 1)):
            result.append('  < leaving group {}'.format(group.get('NAME')))
          else:
            result.append('  keyword {}'.format(name))
          for i in range(abs(n)):
            result.append('    |{0}'.format(inputlines[nline+i]).strip('\n'))
          # a negative n ends the group, so pop the stack
          if (n < 0):
            group = stack.pop(-1)
          bad = False
          found.append(name)
          if ((program in ['GATEWAY', 'SEWARD']) and (name in ['COORD', 'TINKER'])):
            enable_xyz(kw.getparent())
          break
      else:
        n = test_group(inputlines[nline:], kw)
        # n is 1 if a group starts
        if (n > 0):
          stack.append(group)
          group = kw
          result.append('  > entering group {}'.format(name))
          for i in range(abs(n)):
            result.append('    |{0}'.format(inputlines[nline+i].strip('\n')))
          bad = False
          found.append(name)
          if ((program in ['GATEWAY', 'SEWARD']) and (name == 'GROMACS')):
            enable_xyz(kw.getparent())
          break
        # otherwise, no line should be consumed
        elif (n < 1):
          n = 0
    # if nothing has matched, it's either an unknown keyword or syntax error
    else:
      # consume one line
      n = 1
      # print message only for first one of consecutive bad lines
      if (not bad):
        if (anykw):
          result.append('  *** syntax error at "{}" ***'.format(inputlines[nline].strip()))
        else:
          result.append('  *** unknown keyword at "{}" ***'.format(inputlines[nline].strip()))
      result.append('    *** {0}'.format(inputlines[nline].strip('\n')))
      bad = True
      rc = 1
    nline += abs(n)

  # Check for required keywords
  for kw in module.xpath('.//KEYWORD[@INPUT="REQUIRED"]'):
    # skip keywords imported from another module
    if (kw.get('MODULE') != module.get('NAME')):
      continue
    name = kw.get('NAME')
    # skip disabled keywords
    if (name[0] in ['*', '#']):
      continue
    # special cases
    if ((name == 'COORD') and any([i in found for i in ['BASIS (NATIVE)', 'XBAS', 'GROMACS', 'TINKER']])):
      continue
    if ((name == 'BASIS (NATIVE)') and ('XBAS' in found)):
      continue
    if (name not in found):
      result.append('*** Keyword {} is required, but was not found'.format(name))
      rc = 5

  # Check select groups
  for group in module.xpath('.//SELECT'):
    names = []
    for kw in group.xpath('.//KEYWORD'):
      if (kw.get('NAME') in found):
        names.append(kw.get('NAME'))
    if (len(names) > 1):
      result.append('*** Keywords {} are not compatible'.format(', '.join(names)))
      rc = 6

  # Check keyword-specific required/exclusive keywords
  for name in found:
    kw = module.find('.//KEYWORD[@NAME="{}"]'.format(name))
    try:
      req = kw.get('REQUIRE').split(',')
      for r in req:
        alt = r.split('.OR.')
        if (not any([i in found for i in alt])):
          result.append('*** Keyword {} is required by {}, but was not found'.format('|'.join(alt), name))
          rc = 7
    except AttributeError:
      pass
    try:
      exc = kw.get('EXCLUSIVE').split(',')
      for i in exc:
        if (i in found):
          result.append('*** Keyword {} is not compatible with {}'.format(i, name))
          rc = 8
    except AttributeError:
      pass

  # Undo possible changes to keyword names
  for kw in module.xpath('KEYWORD[starts-with(@NAME, "*")] | KEYWORD[starts-with(@NAME, "#")]'):
    kw.set('NAME', '{}'.format(kw.get('NAME')[1:]))

  return (rc, result)



# -*- coding: utf-8 -*-
#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015-2018, Ignacio Fdez. Galván                        *
#***********************************************************************
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os

sys.dont_write_bytecode = True

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

sys.path.append(os.path.abspath('../extensions'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.8'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.mathjax',
    #'sphinx.ext.pngmath',
    'sphinxcontrib.bibtex',
    #'rst2pdf.pdfbuilder',
    'transforms',
    'xmldoc',
    'extractfile',
    'float',
    'molcasbib',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Molcas'
copyright = u'2017–2020, MOLCAS Team'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = os.environ.get('OPENMOLCAS_VERSION', 'unknown')
# The full version, including alpha/beta/rc tags.
release = os.environ.get('OPENMOLCAS_RELEASE', 'unknown')

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['tutorials/tut_alaska.rst']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'molcas'
html_theme_options = {
  'pdf_file': os.environ.get('PDF_FILE', '')
#    'rightsidebar': 'false',
#    'stickysidebar': 'true'
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = { }

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ['_theme']

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = 'Molcas Manual (version {})'.format(version)

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = '<span class="sc">Molcas</span> Manual (version {})'.format(release)

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Molcasdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',

# Maketitle command
  'maketitle': '\maketitle',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, 'Manual.tex', r'Molcas Manual\\(version {})'.format(release), u'MOLCAS Team', 'memoir'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = True

# If true, show page references after internal links.
#latex_show_pagerefs = True

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'molcas', u'Molcas Documentation',
     [u'MOLCAS Team'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  (master_doc, 'Molcas', u'Molcas Documentation',
   u'MOLCAS Team', 'Molcas', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# -- Options for PDF output --------------------------------------------------

# Grouping the document tree into PDF files. List of tuples
# (source start file, target name, title, author, options).
#
# If there is more than one author, separate them with \\.
# For example: r'Guido van Rossum\\Fred L. Drake, Jr., editor'
#
# The options element is a dictionary that lets you override
# this config per-document.
# For example,
# ('index', u'MyProject', u'My Project', u'Author Name',
#  dict(pdf_compressed = True))
# would mean that specific document would be compressed
# regardless of the global pdf_compressed setting.

pdf_documents = [
    ('manual', 'Molcas', u'Molcas Documentation', u'MOLCAS Team', None),
]

# A comma-separated list of custom stylesheets. Example:
#pdf_stylesheets = ['sphinx','kerning','a4']

# A list of folders to search for stylesheets. Example:
#pdf_style_path = ['.', '_styles']

# Create a compressed PDF
# Use True/False or 1/0
# Example: compressed=True
#pdf_compressed = False

# A colon-separated list of folders to search for fonts. Example:
# pdf_font_path = ['/usr/share/fonts', '/usr/share/texmf-dist/fonts/']

# Language to be used for hyphenation support
#pdf_language = "en_US"

# Mode for literal blocks wider than the frame. Can be
# overflow, shrink or truncate
#pdf_fit_mode = "shrink"

# Section level that forces a break page.
# For example: 1 means top-level sections start in a new page
# 0 means disabled
#pdf_break_level = 0

# When a section starts in a new page, force it to be 'even', 'odd',
# or just use 'any'
#pdf_breakside = 'any'

# Insert footnotes where they are defined instead of
# at the end.
#pdf_inline_footnotes = True

# verbosity level. 0 1 or 2
#pdf_verbosity = 0

# If false, no index is generated.
#pdf_use_index = True

# If false, no modindex is generated.
#pdf_use_modindex = True

# If false, no coverpage is generated.
#pdf_use_coverpage = True

# Name of the cover page template to use
#pdf_cover_template = 'sphinxcover.tmpl'

# Documents to append as an appendix to all manuals.
#pdf_appendices = []

# Enable experimental feature to split table cells. Use it
# if you get "DelayedTable too big" errors
#pdf_splittables = False

# Set the default DPI for images
#pdf_default_dpi = 72

# Enable rst2pdf extension modules (default is only vectorpdf)
# you need vectorpdf if you want to use sphinx's graphviz support
#pdf_extensions = ['vectorpdf']

# Page template name for "regular" pages
#pdf_page_template = 'cutePage'

# Show Table Of Contents at the beginning?
#pdf_use_toc = True

# How many levels deep should the table of contents be?
#pdf_toc_depth = 9999

# Add section number to section references
#pdf_use_numbered_links = False

# Background images fitting mode
#pdf_fit_background_mode = 'scale'


# -- Additional options ---------------------------------------------------

primary_domain = ''
highlight_language = 'none'
rst_prolog = open('definitions.sb', 'r').read()
html_compact_lists = True

numfig = True
numfig_secnum_depth = 3
numfig_format = {'section': 'Section %s',
                 'figure': 'Figure %s',
                 'table': 'Table %s',
                 'code-block': 'Block %s'}

latex_toplevel_sectioning = 'part'
latex_elements['papersize'] = 'a4paper'
latex_elements['pointsize'] = '12pt'
latex_elements['fncychap'] = ''
latex_elements['utf8extra'] = r'\DeclareUnicodeCharacter{00A0}{\nobreakspace}'\
                              r'\DeclareUnicodeCharacter{2212}{\textminus}'\
                              r'\DeclareUnicodeCharacter{200B}{\relax}'
latex_elements['fontpkg'] = r'\usepackage[notextcomp]{kpfonts}'
latex_elements['fontenc'] = r'\usepackage[LGR,T1]{fontenc}'
latex_elements['preamble'] = r'''\usepackage{molcas}
\pagestyle{plain}
\setsecnumdepth{subparagraph}
\renewcommand{\thepart}{\arabic{part}}
\numberwithin{chapter}{part}
\setlength\cftchapternumwidth{2em}
\setlength\cftsectionindent{2em}
\setlength\cftsectionnumwidth{3.5em}
\nobibintoc
\renewcommand{\bibsection}{\relax}
\frenchspacing'''

# Fix issues with sphinx.sty
latex_elements['preamble'] += r'''
% These should be done after loading hyperref
\makeatletter%
\@addtoreset{figure}{subsection}%
\@addtoreset{table}{subsection}%
\@addtoreset{literalblock}{subsection}%
\ifspx@opt@mathnumfig%
  \@addtoreset{equation}{subsection}%
\fi%
\makeatother%
% Add part to numbers
\let\oldthefigure\thefigure%
\renewcommand{\thefigure}{\thepart.\oldthefigure}%
\let\oldthetable\thetable%
\renewcommand{\thetable}{\thepart.\oldthetable}%
\let\oldtheliteralblock\theliteralblock%
\renewcommand{\theliteralblock}{\thepart.\oldtheliteralblock}%
\let\oldtheequation\theequation%
\renewcommand{\theequation}{\thepart.\oldtheequation}%
% Missing unicode character(s)?
\DeclareUnicodeCharacter{03A6}{$\Phi$}
'''

latex_additional_files = [ '_latex/molcas.sty' ]

pngmath_dvipng_args = ['-Q', '10']
pngmath_use_preview = True
pngmath_latex_preamble = r'''\usepackage{amsmath,mhchem,kpfonts}
\DeclareMathOperator{\sign}{sign}
\DeclareMathOperator{\abs}{abs}
\DeclareMathOperator{\Tr}{Tr}
\newcommand{\mat}[1]{\boldsymbol{#1}}
\newcommand{\bra}[1]{\left<#1\right|}
\newcommand{\ket}[1]{\left|#1\right>}
\newcommand{\braket}[2]{\left<#1\middle|#2\right>}
\newcommand{\braopket}[3]{\left<#1\middle|#2\middle|#3\right>}'''

# xmldoc extension:
data_dir = 'data'

# extractfile extension:
extract_dir = 'samples'

# references extension:
ref_file = 'references'

def setup(app):
    app.add_css_file('fonts.css')
    app.add_css_file('style.css')
    app.add_css_file('colors.css', title='Default')
    app.add_css_file('nocolors.css', title='No colors', rel='alternate stylesheet')
    app.add_css_file('specific.css')
    app.add_js_file('styleswitcher.js')
    app.add_js_file('functions.js')
    app.add_js_file('mathjax_config.js')
